#!/usr/bin/env python

'''
Created on Dec 6, 2013

:author: jzupka, astepano
:contact: Andrei Stepanov <astepano@redhat.com>
'''

import os
import sys
import imp
import importlib
import select
import time
import stat
import gc
import logging
import traceback
import subprocess
import string
import random
import shutil
import signal
import tempfile

if __name__ == '__main__':
    remote_interface = importlib.import_module('remote_interface')
    ms = importlib.import_module('messenger')
else:
    from virttest.remote_commander import remote_interface
    from virttest.remote_commander import messenger as ms


from six.moves import input


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)
logfile = tempfile.mktemp(suffix='.log', prefix='remote_runner', dir="/tmp")
handler = logging.FileHandler(logfile, "w", encoding=None, delay="true")
logger.addHandler(handler)


def daemonize(pipe_root_path="/tmp"):
    """
    Init daemon.

    :param pipe_root_path: path to directory for pipe.
    :return: [True if child, stdin_path, stdou_path, stderr_path]
    """
    def is_file_open(path):
        """
        Determine process which open file.

        :param path: Path to file.
        :return: [[pid,mode], ... ].
        """
        opens = []
        pids = os.listdir('/proc')
        for pid in sorted(pids):
            try:
                int(pid)
            except ValueError:
                continue
            fd_dir = os.path.join('/proc', pid, 'fd')
            try:
                for filepath in os.listdir(fd_dir):
                    try:
                        p = os.path.join(fd_dir, filepath)
                        link = os.readlink(os.path.join(fd_dir, filepath))
                        if link == path:
                            mode = os.lstat(p).st_mode
                            opens.append([pid, mode])
                    except OSError:
                        continue
            except OSError as e:
                if e.errno == 2:
                    continue
                raise
        return opens

    def daemonize():
        """
        Run guest as a daemon.
        """
        gc_was_enabled = gc.isenabled()
        # Disable gc to avoid bug where gc -> file_dealloc ->
        # write to stderr -> hang.  http://bugs.python.org/issue1336
        gc.disable()
        try:
            pid = os.fork()
            if gc_was_enabled:
                gc.enable()
            if pid > 0:  # If parent return False
                os.waitpid(pid, 0)
                return 0
        except OSError as e:
            sys.stderr.write("Daemonize failed: %s\n" % (e))
            sys.exit(1)

        os.chdir("/")
        os.setsid()
        os.umask(0)

        try:
            pid = os.fork()
            if gc_was_enabled:
                gc.enable()
            if pid > 0:  # If parent Exit
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("Daemonize failed: %s\n" % (e))
            sys.exit(1)

        if gc_was_enabled:
            gc.enable()

        sys.stdout.flush()
        sys.stderr.flush()
        return 1

    stdin_path = os.path.join(pipe_root_path, "stdin")
    stdout_path = os.path.join(pipe_root_path, "stdout")
    stderr_path = os.path.join(pipe_root_path, "stderr")
    results_path = os.path.join(pipe_root_path, "results")
    inputs_path = os.path.join(pipe_root_path, "inputs")

    for f in [stdin_path, stdout_path, stderr_path, results_path, inputs_path]:
        try:
            os.mkfifo(f)
        except OSError as e:
            if e.errno == 17:
                pass

    # Check for a pidfile to see if the daemon already runs
    openers = is_file_open(stdout_path)
    rundaemon = False
    if len(openers) > 0:
        for i in openers:
            if i[1] & stat.S_IWUSR:
                rundaemon = True
                openers.remove(i)
        if len(openers) > 0:
            for i in openers:
                os.kill(int(i[0]), 9)
    time.sleep(0.3)

    # Start the daemon
    child = False
    if not rundaemon:
        child = daemonize()

    if child == 0:
        return (child,
                inputs_path,
                results_path,
                stdin_path,
                stdout_path,
                stderr_path)
    else:
        signal.signal(signal.SIGIO, signal.SIG_DFL)
        return (child,
                results_path,
                inputs_path,
                stdin_path,
                stdout_path,
                stderr_path)


def create_process_cmd():
    """
    Create child process without clean process data thanks that it is possible
    call function and classes from child process.
    """
    r_c, w_p = os.pipe()
    r_p, w_c = os.pipe()
    r_si, w_si = os.pipe()
    r_so, w_so = os.pipe()
    r_se, w_se = os.pipe()
    gc_was_enabled = gc.isenabled()
    # Disable gc to avoid bug where gc -> file_dealloc ->
    # write to stderr -> hang.  http://bugs.python.org/issue1336
    gc.disable()
    pid = os.fork()
    if pid == 0:  # Child process
        os.close(r_p)
        os.close(w_p)
        os.close(w_si)
        os.close(r_so)
        os.close(r_se)
        sys.stdin.close()
        sys.stdout.close()
        sys.stderr.close()
        sys.stdin = os.fdopen(r_si, 'r', 0)
        sys.stdout = os.fdopen(w_so, 'w', 0)
        sys.stderr = os.fdopen(w_se, 'w', 0)
        if gc_was_enabled:
            gc.enable()
        return (0, r_c, w_c, None, None, None)
    else:
        os.close(r_c)
        os.close(w_c)
        os.close(r_si)
        os.close(w_so)
        os.close(w_se)
        if gc_was_enabled:
            gc.enable()
        return (pid, r_p, w_p, w_si, r_so, r_se)


def gen_tmp_dir(root_path):
    """
    Try to create tmp dir with special name.
    """
    path = None
    while (path is None or os.path.exists(path)):
        rname = "runner" + "".join(random.sample(string.letters, 4))
        path = os.path.join(root_path, rname)
        try:
            if not os.path.exists(path):
                os.mkdir(path)
                return path
        except Exception:
            continue


def clean_tmp_dir(path):
    """
    Clean up directory.
    """
    shutil.rmtree(path, True)


def sort_fds_event(fds):
    hup = [x[0] for x in fds if x[1] & select.POLLHUP]
    read = [x[0] for x in fds if x[1] & select.POLLIN]
    write = [x[0] for x in fds if x[1] & select.POLLOUT]
    return hup, read, write


def close_unused_fds(fds):
    """
    Close all file descriptors which are not necessary anymore.

    :param fds: file descriptors
    :type fds: builtin.list
    """
    for fd in fds:
        os.close(fd)


class CmdFinish(object):

    """
    Class used for communication with child process. This class
    """
    __slots__ = ["pid"]

    def __init__(self, parent=False):
        if not parent:
            self.pid = os.getpid()
        else:
            self.pid = os.getppid()
        self.pid = self.pid


class CmdSlave(object):

    """
    Representation of BaseCmd on slave side.
    """

    def __init__(self, baseCmd):
        """
        :param baseCmd: basecmd for encapsulation.
        """
        self.basecmd = baseCmd
        self.cmd_id = baseCmd.cmd_id
        self.obj = None
        self.pid = None
        self.r_pipe = None
        self.w_pipe = None
        self.stdin_pipe = None
        self.stdout_pipe = None
        self.stderr_pipe = None
        self.asynchronous = False
        self.nohup = False
        self.manage = False
        self.msg = None

    def close_pipes(self):
        """
        Close command communication pipe.
        """
        if self.r_pipe is not None:
            os.close(self.r_pipe)
        if self.w_pipe is not None:
            os.close(self.w_pipe)
        if self.stdin_pipe is not None:
            os.close(self.stdin_pipe)
        if self.stdout_pipe is not None:
            os.close(self.stdout_pipe)
        if self.stderr_pipe is not None:
            os.close(self.stderr_pipe)

    def parse_func_name(self, func_name, commander):
        """
        Parse name sended from master.

        format: ``["manage|async|nohup| ", "fnname1", "fnname2", ...]``

        :param func_name: Function name
        :param commander: Where to execute the command (remote or local)
        """
        if func_name[0] == "manage":  # start command in main process.
            self.manage = True
            func_name = func_name[1:]
        if func_name[0] == "async":  # start command in new process.
            self.asynchronous = True
            func_name = func_name[1:]
        if func_name[0] == "nohup":  # start command in new daemon process.
            self.nohup = True
            func_name = func_name[1:]
        if func_name[0] == "python_file_run_with_helper" and not self.manage:
            # the `async` and other prefixes are already consumed.
            raise NotImplementedError(
                "The 'python_file_run_with_helper' is only implemented for "
                "'manage' target.  Please use "
                "'commander.manage.python_file_run_with_helper' to run this.")
        if hasattr(commander, func_name[0]):
            obj = getattr(commander, func_name[0])
        elif func_name[0] in commander.globals:
            obj = commander.globals[func_name[0]]
        elif func_name[0] in commander.locals:
            obj = commander.locals[func_name[0]]
        else:
            obj = globals()[func_name[0]]
        if len(func_name) > 1:
            for name in func_name[1:]:
                obj = getattr(obj, name)
        return obj

    def __call__(self, commander):
        """
        Call command cmd(*args, **kargs)
        """
        self.obj = self.parse_func_name(self.basecmd.func, commander)
        if self.manage:  # start command in main process
            self.basecmd.results = self.obj(*self.basecmd.args,
                                            **self.basecmd.kargs)
            self.basecmd._finished = True
            self.finish(commander)
        elif self.asynchronous:  # start command in new process
            self.basecmd.results = self.__call_async__(commander)
            self.basecmd._async = True
        elif self.nohup:   # start command in new daemon process
            if self.basecmd.cmd_hash is None:
                self.basecmd.cmd_hash = gen_tmp_dir("/tmp")
            self.basecmd.results = self.__call_nohup__(commander)
            self.basecmd._async = True
        else:  # start command in new process but wait for input.
            self.basecmd.results = self.__call_async__(commander)

    def __call_async__(self, commander):
        (self.pid, self.r_pipe, self.w_pipe, self.stdin_pipe,
         self.stdout_pipe, self.stderr_pipe) = create_process_cmd()
        if self.pid == 0:  # Child process make commands
            commander._close_cmds_stdios(self)
            self.msg = ms.Messenger(ms.StdIOWrapperIn(self.r_pipe),
                                    ms.StdIOWrapperOut(self.w_pipe))
            try:
                self.basecmd.results = self.obj(*self.basecmd.args,
                                                **self.basecmd.kargs)
            except Exception:
                err_msg = traceback.format_exc()
                self.msg.write_msg(remote_interface.CmdTraceBack(err_msg))
                sys.exit(-1)
            finally:
                self.msg.write_msg(self.basecmd.results)
                self.msg.write_msg(CmdFinish())
            sys.exit(0)
        else:  # Parent process create communication interface to child process
            self.msg = ms.Messenger(ms.StdIOWrapperIn(self.r_pipe),
                                    ms.StdIOWrapperOut(self.w_pipe))

    def __call_nohup__(self, commander):
        (pid, self.r_path, self.w_path, self.stdin_path, self.stdout_path,
         self.stderr_path) = daemonize(self.basecmd.cmd_hash)
        if pid == 1:  # Child process make commands
            commander._close_cmds_stdios(self)
            (self.pid, r_pipe, w_pipe, stdin_pipe,
             stdout_pipe, stderr_pipe) = create_process_cmd()
            if self.pid == 0:  # Child process make commands
                self.msg = ms.Messenger(ms.StdIOWrapperIn(r_pipe),
                                        ms.StdIOWrapperOut(w_pipe))
                try:
                    self.basecmd.results = self.obj(*self.basecmd.args,
                                                    **self.basecmd.kargs)
                except Exception:
                    err_msg = traceback.format_exc()
                    self.msg.write_msg(remote_interface.CmdTraceBack(err_msg))
                    sys.exit(-1)
                finally:
                    self.msg.write_msg(self.basecmd.results)
                sys.exit(0)
            else:
                # helper child process open communication pipes.
                # This process is able to manage problem with connection width
                # main parent process. It allows start unchanged child process.
                self.r_pipe = os.open(self.r_path, os.O_RDONLY)
                self.w_pipe = os.open(self.w_path, os.O_WRONLY)
                sys.stdout = os.fdopen(os.open(self.stdout_path, os.O_WRONLY),
                                       "w",
                                       0)
                sys.stderr = os.fdopen(os.open(self.stderr_path, os.O_WRONLY),
                                       "w",
                                       0)
                sys.stdin = os.fdopen(os.open(self.stdin_path, os.O_RDONLY),
                                      "r",
                                      0)

                w_fds = [r_pipe, w_pipe, stdin_pipe, stdout_pipe, stderr_pipe]
                m_fds = [self.r_pipe,
                         self.w_pipe,
                         sys.stdin.fileno(),
                         sys.stdout.fileno(),
                         sys.stderr.fileno()]
                p = select.poll()
                p.register(r_pipe)
                p.register(w_pipe)
                # p.register(stdin_pipe)
                p.register(stdout_pipe)
                p.register(stderr_pipe)
                p.register(self.r_pipe)
                # p.register(self.w_pipe)
                p.register(sys.stdin.fileno())
                # p.register(sys.stdout.fileno())
                # p.register(sys.stderr.fileno())
                io_map = {r_pipe: self.w_pipe,
                          self.r_pipe: w_pipe,
                          sys.stdin.fileno(): stdin_pipe,
                          stdout_pipe: sys.stdout.fileno(),
                          stderr_pipe: sys.stderr.fileno()}
                while 1:
                    d = p.poll()
                    w_ev = [x for x in d if x[0] in w_fds]
                    m_ev = [x for x in d if x[0] in m_fds]
                    w_hup, w_read, _ = sort_fds_event(w_ev)
                    m_hup, m_read, _ = sort_fds_event(m_ev)
                    if m_hup:
                        time.sleep(0.1)
                    if w_hup:  # child process finished
                        for r in w_read:
                            data = os.read(r, 16384)
                            os.write(io_map[r], data)
                        break
                    for r in w_read:
                        data = os.read(r, 16384)
                        os.write(io_map[r], data)
                    for r in m_read:
                        data = os.read(r, 16384)
                        os.write(io_map[r], data)
                self.msg = ms.Messenger(ms.StdIOWrapperIn(self.r_pipe),
                                        ms.StdIOWrapperOut(self.w_pipe))
                self.msg.write_msg(CmdFinish())
                exit(0)
        else:  # main process open communication named pipes.
            self.w_pipe = os.open(self.w_path, os.O_WRONLY)
            self.r_pipe = os.open(self.r_path, os.O_RDONLY)
            self.stdout_pipe = os.open(self.stdout_path, os.O_RDONLY)
            self.stderr_pipe = os.open(self.stderr_path, os.O_RDONLY)
            self.stdin_pipe = os.open(self.stdin_path, os.O_WRONLY)
            self.msg = ms.Messenger(ms.StdIOWrapperIn(self.r_pipe),
                                    ms.StdIOWrapperOut(self.w_pipe))

    def work(self):
        """
        Wait for message from running child process
        """
        succ, msg = self.msg.read_msg()
        if isinstance(msg, CmdFinish):
            try:
                pid, _ = os.waitpid(msg.pid, 0)
            except OSError:
                pid = msg.pid
            if (succ is False or pid == msg.pid):
                self.basecmd._finished = True
                return True
            else:
                return False
        else:
            self.basecmd.results = msg

    def recover_paths(self):
        """
        Helper function for reconnect to daemon/nohup process.
        """
        self.stdin_path = os.path.join(self.basecmd.cmd_hash, "stdin")
        self.stdout_path = os.path.join(self.basecmd.cmd_hash, "stdout")
        self.stderr_path = os.path.join(self.basecmd.cmd_hash, "stderr")
        self.w_path = os.path.join(self.basecmd.cmd_hash, "results")
        self.r_path = os.path.join(self.basecmd.cmd_hash, "inputs")

    def recover_fds(self):
        """
        Helper function for reconnect to daemon/nohup process.
        """
        if self.r_pipe is None:
            self.recover_paths()
            self.w_pipe = os.open(self.w_path, os.O_WRONLY)
            self.r_pipe = os.open(self.r_path, os.O_RDONLY)
            self.stdin_pipe = os.open(self.stdin_path, os.O_WRONLY)
            self.stdout_pipe = os.open(self.stdout_path, os.O_RDONLY)
            self.stderr_pipe = os.open(self.stderr_path, os.O_RDONLY)
            self.msg = ms.Messenger(ms.StdIOWrapperIn(self.r_pipe),
                                    ms.StdIOWrapperOut(self.w_pipe))

    def finish(self, commander):
        """
        Remove cmd from commander commands on finish of process.
        """
        self.close_pipes()
        if self.basecmd.cmd_hash:
            clean_tmp_dir(self.basecmd.cmd_hash)
            self.basecmd.cmd_hash = None
        del commander.cmds[self.cmd_id]


class CommanderSlave(ms.Messenger):

    """
    Class commander slace is responsible for communication with commander
    master. It invoke commands to slave part and receive messages from them.
    For communication is used only stdin and stdout which are streams from
    slave part.
    """

    def __init__(self, stdin, stdout, o_stdout, o_stderr):
        super(CommanderSlave, self).__init__(stdin, stdout)
        self._exit = False
        self.cmds = {}
        self.globals = {}
        self.locals = {}
        self.o_stdout = o_stdout
        self.o_stderr = o_stderr

    def cmd_loop(self):
        """
        Wait for commands from master and receive results and outputs from
        commands.
        """
        try:
            while (not self._exit):
                stdios = [self.stdin, self.o_stdout, self.o_stderr]
                r_pipes = [cmd.r_pipe for cmd in list(self.cmds.values())
                           if cmd.r_pipe is not None]
                stdouts = [cmd.stdout_pipe for cmd in list(self.cmds.values())
                           if cmd.stdout_pipe is not None]
                stderrs = [cmd.stderr_pipe for cmd in list(self.cmds.values())
                           if cmd.stderr_pipe is not None]

                r, _, _ = select.select(
                    stdios + r_pipes + stdouts + stderrs, [], [])

                if self.stdin in r:  # command from controller
                    m = self.read_msg()
                    if m[0] is False:
                        logger.info("Other side is closed.")
                        break
                    if m[0] is None:
                        logger.info("Reading is timeouted.")
                        break
                    cmd = CmdSlave(m[1])
                    self.cmds[cmd.cmd_id] = cmd
                    try:
                        # There is hidden bug. We can bump into condition when
                        # we running some function-command. That function dump
                        # some data to stdio. But, stdio is out of
                        # consideration at this point. It can stuck.
                        cmd(self)
                        self.write_msg(cmd.basecmd)
                    except Exception:
                        err_msg = traceback.format_exc()
                        self.write_msg(
                            remote_interface.CommanderError(err_msg))

                if self.o_stdout in r:  # Send message from stdout
                    msg = os.read(self.o_stdout, 16384)
                    self.write_msg(remote_interface.StdOut(msg))
                if self.o_stderr in r:  # Send message from stdout
                    msg = os.read(self.o_stderr, 16384)
                    self.write_msg(remote_interface.StdErr(msg))

                # test all commands for io
                for cmd in list(self.cmds.values()):
                    if cmd.stdout_pipe in r:  # command stdout
                        data = os.read(cmd.stdout_pipe, 16384)
                        if data != "":  # pipe is not closed on another side.
                            self.write_msg(remote_interface.StdOut(data,
                                                                   cmd.cmd_id))
                        else:
                            os.close(cmd.stdout_pipe)
                            cmd.stdout_pipe = None
                    if cmd.stderr_pipe in r:  # command stderr
                        data = os.read(cmd.stderr_pipe, 16384)
                        if data != "":  # pipe is not closed on another side.
                            self.write_msg(remote_interface.StdErr(data,
                                                                   cmd.cmd_id))
                        else:
                            os.close(cmd.stderr_pipe)
                            cmd.stderr_pipe = None
                    if cmd.r_pipe in r:  # command results
                        if cmd.work():
                            cmd.finish(self)
                        self.write_msg(cmd.basecmd)
        except Exception:
            err_msg = traceback.format_exc()
            self.write_msg(remote_interface.CommanderError(err_msg))

    def _close_cmds_stdios(self, exclude_cmd):
        for cmd in list(self.cmds.values()):
            if cmd is not exclude_cmd:
                cmd.close_pipes()


class CommanderSlaveCmds(CommanderSlave):

    """
    Class extends CommanderSlave and adds to them special commands like
    shell process, interactive python, send_msg to cmd.
    """

    def __init__(self, stdin, stdout, o_stdout, o_stderr):
        super(CommanderSlaveCmds, self).__init__(stdin, stdout,
                                                 o_stdout, o_stderr)

        while (1):
            succ, data = self.read_msg()
            if succ and data == "start":
                break
        self.write_msg("Started")

    def shell(self, cmd):
        """
        Starts shell process. Stdout is automatically copyed to basecmd.stdout

        :param cmd: Command which should be started.
        :return: basecmd with return code of cmd.
        """
        process = subprocess.Popen(cmd,
                                   shell=True,
                                   stdin=sys.stdin,
                                   stdout=sys.stdout,
                                   stderr=sys.stderr,
                                   universal_newlines=True)

        return process.wait()

    def interactive(self):
        """
        Starts interactive python.
        """
        while 1:
            out = input()
            if out == "":
                return
            try:
                exec(out)
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                print("On Guest exception from: \n" + "".join(
                    traceback.format_exception(exc_type,
                                               exc_value,
                                               exc_traceback)))
                print("FAIL: Guest command exception.")

    def send_msg(self, msg, cmd_id):
        """
        Send msg to cmd with id == cmd_id

        :param msg: message passed to cmd over the stdin
        :type msg: str
        :param cmd_id: id of cmd.
        """
        os.write(self.cmds[cmd_id].stdin_pipe, msg)

    def register_cmd(self, basecmd, basecmd_cmd_id):
        """
        Second side of set_commander cmd from master. It register existing
        cmd to CommandSlave dict.

        :param basecmd: cmd which should be added to CommandSlave dict
        :type basecmd: BaseCmd
        :param basecmd_cmd_id: number under which should be stored
        :type basecmd_cmd_id: int
        """
        remote_interface.BaseCmd.single_cmd_id = basecmd_cmd_id
        cmd = CmdSlave(basecmd)
        self.cmds[basecmd.cmd_id] = cmd
        if cmd.basecmd.cmd_hash is not None:
            cmd.recover_fds()
        return basecmd

    def add_function(self, f_code):
        """
        Adds function to client code.

        :param f_code: Code of function.
        :type f_code: str.
        """
        exec(f_code, globals(), globals())

    def copy_file(self, name, path, content):
        """
        Really naive implementation of copping files. Should be used only for
        short files.
        """
        f = open(os.path.join(path, name), "w")
        f.write(content)
        f.close()

    def import_src(self, name, path=None):
        """
        Import file to running python session.
        """
        if path:
            if path not in sys.path:
                sys.path.append(path)
        mod = __import__(name, globals(), locals())
        globals()[name] = mod
        sys.modules[name] = mod

    def exit(self):
        """
        Method for killing command slave.
        """
        self._exit = True
        return "bye"

    def python_file_run_with_helper(self, test_path):
        """
        Call run() function at external python module.

        :param test_path: Path to python file. Call run() at this module.
        :return: module.run().return
        """
        assert os.path.isfile(test_path)
        module = imp.load_source('ext_test', test_path)
        assert hasattr(module, "run")
        run = getattr(module, "run")
        helper = Helper(self)
        return run(helper)


class Helper(object):
    """Passed to external test on VM."""
    __slots__ = ["messenger"]

    def __init__(self, messenger):
        self.messenger = messenger

    def info(self, *args, **kargs):
        logger.info(*args, **kargs)
        self.flush_buf()

    def flush_buf(self):
        msg = self.messenger
        stdios = [msg.o_stdout, msg.o_stderr]
        r, _, _ = select.select(stdios, [], [], 0)  # Not blocking.
        if msg.o_stdout in r:  # Send message from stdout
            data = os.read(msg.o_stdout, 16384)
            msg.write_msg(remote_interface.StdOut(data))
        if msg.o_stderr in r:  # Send message from stdout
            data = os.read(msg.o_stderr, 16384)
            msg.write_msg(remote_interface.StdErr(data))

    def query_master(self, *args, **kargs):
        """Read CmdRespond from master."""
        # First of all, dump stdio to master. Otherwise it can block on
        # logging.info() function.
        self.flush_buf()
        # Sent actual request.
        cmd = remote_interface.CmdQuery(*args, **kargs)
        msg = self.messenger
        msg.write_msg(cmd)
        succ, data = msg.read_msg()
        assert succ
        assert isinstance(data, remote_interface.CmdRespond)
        return data.respond


def remote_agent(in_stream_cls, out_stream_cls):
    """
    Connect file descriptors to right pipe and start slave command loop.
    When something happend it raise exception which could be caught by cmd
    master.

    :params in_stream_cls: Class encapsulated input stream.
    :params out_stream_cls: Class encapsulated output stream.
    """
    try:
        sys.stdout.flush()
        sys.stderr.flush()
        fd_stdin = sys.stdin.fileno()    # == 0
        fd_stdout = sys.stdout.fileno()  # == 1
        fd_stderr = sys.stderr.fileno()  # == 2
        orig_stdout = os.dup(fd_stdout)  # Original stdout to master.
        orig_stderr = os.dup(fd_stderr)  # Original srderr to master.
        soutr, soutw = os.pipe()
        serrr, serrw = os.pipe()
        os.dup2(soutw, fd_stdout)        # Stdout == pipe
        os.dup2(serrw, fd_stderr)        # Stderr == pipe
        os.close(soutw)  # Close pipe as it stays opened in stdout.
        os.close(serrw)  # Close pipe as it stays opened in stderr.
        sys.stdout = os.fdopen(fd_stdout, 'w', 0)
        sys.stderr = os.fdopen(fd_stderr, 'w', 0)
        os.write(orig_stdout, "#")

        # Logging goes to the pipe.
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        #logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

        w_stdin = None
        w_stdout = out_stream_cls(orig_stdout)
        w_stdin = in_stream_cls(fd_stdin)

        cmd = CommanderSlaveCmds(w_stdin,
                                 w_stdout,
                                 soutr,
                                 serrr)

        cmd.cmd_loop()
    except SystemExit:
        pass
    except Exception:
        e = traceback.format_exc()
        sys.stderr.write(e)
        # traceback.print_exc()


if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1] == "agent":
            remote_agent(ms.StdIOWrapperIn, ms.StdIOWrapperOut)
        elif sys.argv[1] == "agent_base64":
            remote_agent(ms.StdIOWrapperInBase64, ms.StdIOWrapperOutBase64)



"""
Classes to support XML for channel devices

http://libvirt.org/formatdomain.html#elementCharSerial
"""

from virttest.libvirt_xml import base
from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices.character import CharacterBase


class Channel(CharacterBase):

    __slots__ = ('source', 'target', 'alias', 'address')

    def __init__(self, type_name='unix', virsh_instance=base.virsh):
        accessors.XMLElementDict('source', self, parent_xpath='/',
                                 tag_name='source')
        accessors.XMLElementDict('target', self, parent_xpath='/',
                                 tag_name='target')
        # example for new slots :  alias and address
        #
        # <?xml version='1.0' encoding='UTF-8'?>
        # <channel type="pty">
        #   <source path="/dev/pts/10" />
        #   <target name="pty" type="virtio" />
        #   <alias name="pty" />
        #   <address bus="0" controller="0" type="virtio-serial" />
        # </channel>
        accessors.XMLElementDict('alias', self, parent_xpath='/',
                                 tag_name='alias')
        accessors.XMLElementDict('address', self, parent_xpath='/',
                                 tag_name='address')
        super(
            Channel, self).__init__(device_tag='channel', type_name=type_name,
                                    virsh_instance=virsh_instance)



import os
import logging
import signal
import threading
try:
    import queue as Queue
except ImportError:
    import Queue

import aexpect

from enum import Enum
from virttest import utils_misc


class GDBError(Exception):

    """
    General module exception class
    """

    pass


class GDBCmdError(GDBError):

    """
    Exception raised when calling an gdb command.
    """

    def __init__(self, command, msg):
        self.command = command
        self.msg = msg

    def __str__(self):
        return 'Error when calling GDB command %s:\n%s' % (
            self.command, self.msg)


def _split_result_str(result_str):
    """
    Helper function to split a string into a list of
    (key, value) tuples while minding braces matching.
    """
    result = []
    key = ''
    value = ''
    current = 'key'
    braces = []
    for ch in result_str:
        if ch == '=' and not braces:
            if current == 'key':
                current = 'value'
        elif ch == ',' and not braces:
            result.append((key, value))
            key = ''
            value = ''
            if current == 'value':
                current = 'key'
        else:
            if ch in ['{', '[']:
                braces.append(ch)
            elif ch == ']':
                if braces[-1] == '[':
                    braces.pop()
            elif ch == '}':
                if braces[-1] == '{':
                    braces.pop()

            if current == 'key':
                key += ch
            elif current == 'value':
                value += ch

    if not braces:
        result.append((key, value))
    return result


def _parse_result(result_str):
    """
    Helper function to parse a GDB/MI result string into Python
    friendly data collections like lists and dicts.
    """
    if result_str.startswith('"') and result_str.endswith('"'):
        return result_str[1:-1]

    result = {}
    if result_str.startswith('[') and result_str.endswith(']'):
        result_str = result_str[1:-1]
        result = []
    elif result_str.startswith('{') and result_str.endswith('}'):
        result_str = result_str[1:-1]

    if not result_str:
        return result

    for key, value_str in _split_result_str(result_str):
        value = _parse_result(value_str)
        if type(result) == list:
            result.append((key, value))
        elif type(result) == dict:
            result[key] = value

    return result


class GDB(aexpect.Expect):

    """
    Class to manipulate a inferior process in gdb.
    """

    def __init__(self, command=None):
        self.running = False
        self.terminated = False
        self.exiting = False
        self.cmd_lock = threading.Lock()
        self.inferior_command = command

        self.output_queue = Queue.Queue()
        self.cmd_queue = Queue.Queue()
        self.cmd_result_queue = Queue.Queue()

        self.notify_async_output = []
        self.exec_async_output = []
        self.status_async_output = []
        self.log_stream = []
        self.console_stream = []
        self.target_stream = []
        self.callback_threads = []
        self.prompt = '(gdb)'
        self.pid = None

        dummy_cb = (lambda self, info, params: None, None)
        self.callbacks = {
            "stop": dummy_cb,
            "start": dummy_cb,
            "termination": dummy_cb,
            "break": dummy_cb,
            "signal": dummy_cb,
        }

        self.thread_groups = {}

        command = 'gdb --quiet --interpreter=mi %s' % command
        aexpect.Expect.__init__(
            self,
            command=command,
        )
        self.log_polling_thread = threading.Thread(target=self._poll_log)
        self.log_polling_thread.start()
        self._read_output_until_prompt(timeout=3)

    def _parse_notify_async_line(self, line):
        """
        Parse a GDB/MI async notify line and change state accordingly.

        :param line: GDB/MI line to be parsed.
        """
        event, info_str = line.lstrip('=').split(',', 1)
        info = _parse_result(info_str)
        if event == 'thread-group-added':
            self.thread_groups[info['id']] = {
                'status': 'stopped',
                'pid': None,
                'threads': set()
            }
            self.current_thread_group = info['id']
        elif event == 'thread-group-started':
            pid = int(info['pid'])
            t_group = self.thread_groups[info['id']]
            t_group['status'] = 'running'
            t_group['pid'] = self.pid = pid
        elif event == 'thread-group-exited':
            t_group = self.thread_groups[info['id']]
            t_group['status'] = 'stopped'
            t_group['pid'] = self.pid = None
        elif event == 'thread-created':
            t_group = self.thread_groups[info['group-id']]
            t_id = int(info['id'])
            t_group['threads'].add(t_id)
        elif event == 'thread-exited':
            t_group = self.thread_groups[info['group-id']]
            t_id = int(info['id'])
            t_group['threads'].remove(t_id)
        elif event in ['library-loaded', 'library-unloaded']:
            pass
        else:
            logging.warning('Unprocessed gdb async notification:\n%s', line)

    def _parse_exec_async_line(self, line):
        """
        Parse a GDB/MI async exec line and change state accordingly.

        :param line: GDB/MI line to be parsed.
        """
        event, info_str = line.split(',', 1)
        info = _parse_result(info_str)
        if event == '*stopped':
            self.running = False
            self._callback('stop', info)
            if info:
                if 'reason' in info:
                    if info['reason'] == 'breakpoint-hit':
                        self._callback('break', info)
                    if info['reason'] == 'signal-received':
                        self._callback('signal', info)
                    if info['reason'] in ['exited', 'exited-normally',
                                          'exited-signalled']:
                        self.terminated = True
                        self._callback('termination', info)
                else:
                    for key in info:
                        logging.warning('Stopped without reason')
                        logging.warning('%s: %s', key, info[key])
            else:
                self._callback('termination', info)
        if event == '*running':
            if not self.running:
                self.running = True
                self._callback('start', info)

    def _parse_status_async_line(self, line):
        """
        Parse a GDB/MI async status line and change state accordingly.

        :param line: GDB/MI line to be parsed.
        """
        pass

    def _parse_cmd_result_line(self, line):
        """
        Parse a GDB/MI command result line and change state accordingly.

        :param line: GDB/MI line to be parsed.
        """
        command = self.cmd_queue.get()
        result = {'command': command}
        res = line.lstrip('^').split(',', 1)
        result['status'] = res[0]
        if len(res) == 2:
            result['info'] = _parse_result(res[1])
        elif len(res) == 1:
            result['info'] = {}
        self.cmd_result_queue.put(result)

    def _read_output_until_prompt(self, timeout=300):
        """
        Read GDB/MI output until a command prompt reached.

        :param timeout: Max time for the reading.
        """
        result = []
        while True:
            line = self.output_queue.get(timeout=timeout)
            result.append(line)
            if line.startswith('(gdb)'):
                return result

    def _poll_log(self):
        """
        Read GDB/MI output continuously and parse lines according to its
        inital until `exit` variable is set.
        """
        while True:
            res = self.read_nonblocking()
            if res.strip():
                for line in res.splitlines():
                    if line.startswith('='):
                        self._parse_notify_async_line(line)
                    elif line.startswith('*'):
                        self._parse_exec_async_line(line)
                    elif line.startswith('+'):
                        self._parse_status_async_line(line)
                    elif line.startswith('^'):
                        self._parse_cmd_result_line(line)
                    else:
                        self.output_queue.put(line)
            if self.exiting:
                break

    def _callback(self, callback_type, info):
        """
        General callback function to call specific type of callback funtions.

        :param callback_type: Could be one of "stop", "start", "termination",
                              "break" or "signal"
        """
        callback_func, params = self.callbacks[callback_type]
        logging.debug('gdb is Calling back %s' % callback_type)
        thread = threading.Thread(
            target=callback_func,
            args=(self, info, params),
            name=callback_type
        )
        thread.start()
        self.callback_threads.append(thread)

    def set_callback(self, callback_type, func, params=None):
        """
        Set a callback function to a customized function.

        :param callback_type: Could be one of "stop", "start", "termination",
                              "break" or "signal"
        :param func: Function to be set as callback
        :param params: Parameters to be passed to callback function
        """
        self.callbacks[callback_type] = (func, params)

    def stop(self):
        """
        Stop inferior by sending a SIGINT signal.
        """
        if self.running:
            self.send_signal('SIGINT')
            self.wait_for_stop()

    def kill(self):
        """
        Kill inferior by sending a SIGTERM signal.
        """
        def temp_callback(gdb, info, params):
            if info['signal-name'] == 'SIGTERM':
                self.cont()

        if self.running:
            stop_cb, stop_cb_params = self.callbacks['signal']
            self.set_callback('signal', temp_callback)
            self.send_signal('SIGTERM')
            self.wait_for_termination()
            self.set_callback('signal', stop_cb, stop_cb_params)

    def cont(self):
        """
        Continue a stopped inferior.
        """
        self.cmd('-exec-continue')
        if not self.running and not self.terminated:
            self.wait_for_start()

    def insert_break(self, break_func):
        """
        Insert a function breakpoint.

        :param break_func: Function at which breakpoint inserted
        """
        return self.cmd('-break-insert -f %s' % break_func)

    def back_trace(self):
        """
        Get current backtrace stack as a list of lines.
        """
        result = self.cmd('-stack-list-frames')

        bts = []
        for line in result['info']['stack']:
            bt = line[1]
            bt_line = "#%2s" % bt['level']
            if 'addr' in bt:
                bt_line += " %s" % bt['addr']
            if 'func' in bt:
                bt_line += " %-30s" % bt['func']
            if 'fullname' in bt:
                bt_line += " %s:%s" % (bt['fullname'], bt['line'])
            elif 'from' in bt:
                bt_line += " %s" % bt['from']
            bts.append(bt_line)
        return bts

    def run(self, arg_str=''):
        """
        Start the inferior with an optional argument string.

        :param arg_str: Argument the inferior to be called with
        """
        if not self.running:
            cmd_line = '-exec-run ' + arg_str
            result = self.cmd(cmd_line)
            self.wait_for_start()
            self.terminated = False
            return result

    def send_signal(self, signal_name):
        """
        Send a signal to the inferior.

        :param signal_name: Signal name as a string or integer
        """
        if isinstance(signal_name, str):
            signal_name = getattr(signal, signal_name)

        if isinstance(signal_name, Enum):
            signal_name = signal_name.value

        if isinstance(signal_name, int):
            os.kill(self.pid, signal_name)
        else:
            raise ValueError("Signal should be a string or an integer.")

    def cmd(self, command, cont=True):
        """
        Call a gdb of GDB/MI command.

        :param command: Command line to be called
        :param cont: Whether continue the inferior after calling the command
        """
        was_running = False
        self.cmd_lock.acquire()
        try:
            if self.running:
                was_running = True
                self.stop()
            self.cmd_queue.put(command)
            self.sendline(command)
            result = self.cmd_result_queue.get()
            if result['command'] != command:
                raise GDBCmdError('Mismatch command result line got %s but '
                                  'expecting %s' %
                                  (result['command'], command))
            if result['status'] == 'exit':
                self.exiting = True
            elif result['status'] == 'error':
                raise GDBCmdError(command, result['info']['msg'])
            elif result['status'] == 'running':
                pass
            elif result['status'] == 'connected':
                pass
        finally:
            self.cmd_lock.release()
            if was_running and cont:
                self.cont()
        return result

    def wait_for_start(self, timeout=60):
        """
        Wait the inferior to start.

        :param timeout: Max time to wait
        """
        logging.debug("Waiting for gdb inferior %s to start"
                      % self.inferior_command)
        return utils_misc.wait_for(
            lambda: self.running,
            timeout,
            step=0.1,
        )

    def wait_for_stop(self, timeout=60):
        """
        Wait the inferior to be stopped.

        :param timeout: Max time to wait
        """
        logging.debug("Waiting for gdb inferior %s to stop"
                      % self.inferior_command)
        res = utils_misc.wait_for(
            lambda: not self.running,
            timeout,
            step=0.1,
        )
        return res

    def wait_for_termination(self, timeout=60):
        """
        Wait the gdb session to be exited.

        :param timeout: Max time to wait
        """
        logging.debug("Waiting for gdb to terminate")
        return utils_misc.wait_for(
            lambda: self.terminated,
            timeout,
            step=0.1,
        )

    def exit(self):
        """
        Exit the gdb session.
        """
        self.cmd('-gdb-exit', cont=False)
        self.log_polling_thread.join()
        for thread in self.callback_threads:
            thread.join()



"""
oVirt SDK wrapper module.

:copyright: 2008-2012 Red Hat Inc.
"""


import time
import logging

import ovirtsdk4 as sdk
import ovirtsdk4.types as types

from virttest import virt_vm


_api = None
_connected = False


class WaitStateTimeoutError(Exception):

    def __init__(self, msg, output):
        Exception.__init__(self, msg, output)
        self.msg = msg
        self.output = output


class WaitVMStateTimeoutError(WaitStateTimeoutError):

    def __str__(self):
        str = "Timeout expired when waiting for VM to %s,"
        str += " actual state is: %s"
        return str % (self.msg, self.output)


class WaitHostStateTimeoutError(WaitStateTimeoutError):

    def __str__(self):
        str = "Timeout expired when waiting for Host to %s,"
        str += " actual state is: %s"
        return str % (self.msg, self.output)


def connect(params):
    """
    Connect ovirt manager API.
    """
    url = params.get('ovirt_engine_url')
    username = params.get('ovirt_engine_user')
    password = params.get('ovirt_engine_password')
    version = params.get('ovirt_engine_version')

    if not all([url, username, password]):
        logging.error('ovirt_engine[url|user|password] are necessary!!')

    if version is None:
        version = types.Version(major='4', minor='2')
    else:
        version = types.Version(version)

    global connection, _connected

    try:
        # Try to connect oVirt API if connection doesn't exist,
        # otherwise, directly return existing API connection.
        if not _connected:
            connection = sdk.Connection(
                url=url,
                username=username,
                password=password,
                insecure=True
            )
            _connected = True
            return connection, version
        else:
            return connection, version
    except Exception as e:
        logging.error('Failed to connect: %s\n' % str(e))
    else:
        logging.info('Succeed to connect oVirt/Rhevm manager\n')


def disconnect():
    """
    Disconnect ovirt manager connection.
    """
    global connection, _connected

    if _connected:
        return connection.close()


class VMManager(virt_vm.BaseVM):

    """
    This class handles all basic VM operations for oVirt.
    """

    def __init__(self, name, params, root_dir=None, address_cache=None,
                 state=None):
        """
        Initialize the object and set a few attributes.

        :param name: The name of the object
        :param params: A dict containing VM params (see method
                       make_create_command for a full description)
        :param root_dir: Base directory for relative filenames
        :param address_cache: A dict that maps MAC addresses to IP addresses
        :param state: If provided, use this as self.__dict__
        """

        if state:
            self.__dict__ = state
        else:
            self.process = None
            self.serial_console = None
            self.redirs = {}
            self.vnc_port = 5900
            self.vnclisten = "0.0.0.0"
            self.pci_assignable = None
            self.netdev_id = []
            self.device_id = []
            self.pci_devices = []
            self.uuid = None
            self.only_pty = False
            self.remote_sessions = []

        self.spice_port = 8000
        self.name = name
        self.params = params
        self.root_dir = root_dir
        self.address_cache = address_cache
        self.vnclisten = "0.0.0.0"
        self.driver_type = "v2v"

        super(VMManager, self).__init__(self.name, params)
        (self.connection, self.version) = connect(params)

        if self.name:
            self.update_instance()

    def update_instance(self):
        vms_service = self.connection.system_service().vms_service()
        vms = vms_service.list(search='name=%s' % self.name)
        if vms:
            vm = vms[0]
            self.instance = vms_service.vm_service(vm.id)
        else:
            self.instance = None

    def list(self):
        """
        List all of VMs.
        """
        vm_list = []
        try:
            vms = self.connection.system_service().vms_service().list()
            for i in range(len(vms)):
                vm_list.append(vms[i].name)
            return vm_list
        except Exception as e:
            logging.error('Failed to get vms:\n%s' % str(e))

    def state(self):
        """
        Return VM state.
        """
        try:
            self.update_instance()
            return self.instance.get().status
        except Exception as e:
            logging.error('Failed to get %s status:\n%s' % (self.name, str(e)))

    def get_mac_address(self, net_name='*'):
        """
        Return MAC address of a VM.
        """
        if not self.instance:
            self.update_instance()
        vnet_list = self.instance.nics_service().list()
        if net_name != '*':
            vnet_list = [vnet for vnet in vnet_list if vnet.name == net_name]
        try:
            return [vnet.mac.address for vnet in vnet_list if vnet.mac.address]
        except Exception as e:
            logging.error('Failed to get %s status:\n%s' % (self.name, str(e)))

    def lookup_by_storagedomains(self, storage_name):
        """
        Lookup VM object in storage domain according to VM name.
        """
        try:
            sds_service = self.connection.system_service().storage_domains_service()
            export_sd = sds_service.list(search=storage_name)[0]
            export_vms_service = sds_service.storage_domain_service(export_sd.id).vms_service()
            target_vm = [vm for vm in export_vms_service.list() if vm.name == self.name][0]
            return target_vm
        except Exception as e:
            logging.error('Failed to get %s from %s:\n%s' % (self.name,
                                                             storage_name, str(e)))

    def is_dead(self):
        """
        Judge if a VM is dead.
        """
        if self.state() == types.VmStatus.DOWN:
            logging.info('VM %s status is <Down>' % self.name)
            return True
        else:
            return False

    def is_alive(self):
        """
        Judge if a VM is alive.
        """
        return not self.is_dead()

    def is_paused(self):
        """
        Return if VM is suspend.
        """
        if self.state() == types.VmStatus.SUSPENDED:
            return True
        else:
            logging.debug('VM %s status is %s ' % (self.name, self.state()))
            return False

    def start(self, wait_for_up=True, timeout=300):
        """
        Start a VM.
        """
        end_time = time.time() + timeout
        if self.is_dead():
            logging.info('Starting VM %s' % self.name)
            self.instance.start()
            vm_powering_up = False
            vm_up = False
            while time.time() < end_time:
                if self.state() == types.VmStatus.POWERING_UP:
                    vm_powering_up = True
                    if wait_for_up:
                        logging.info('Waiting for VM to reach <Up> status')
                        if self.state() == types.VmStatus.UP:
                            vm_up = True
                            break
                    else:
                        break
                elif self.state() == types.VmStatus.UP:
                    vm_up = True
                    break
                time.sleep(1)
            if not vm_powering_up and not vm_up:
                raise WaitVMStateTimeoutError("START", self.state())
        else:
            logging.debug('VM is alive')

    def suspend(self, timeout):
        """
        Suspend a VM.
        """
        end_time = time.time() + timeout
        vm_suspend = False
        while time.time() < end_time:
            try:
                logging.info('Suspend VM %s' % self.name)
                self.instance.suspend()
                logging.info('Waiting for VM to reach <Suspended> status')
                if self.is_paused():
                    vm_suspend = True
                    break
            except Exception as e:
                if e.reason == 'Bad Request' \
                        and 'asynchronous running tasks' in e.detail:
                    logging.warning("VM has asynchronous running tasks, "
                                    "trying again")
                    time.sleep(1)
                else:
                    raise e
            time.sleep(1)
        if not vm_suspend:
            raise WaitVMStateTimeoutError("SUSPEND", self.state())

    def resume(self, timeout):
        """
        Resume a suspended VM.
        """
        end_time = time.time() + timeout
        try:
            if self.state() != 'up':
                logging.info('Resume VM %s' % self.name)
                self.instance.start()
                logging.info('Waiting for VM to <UP> status')
                vm_resume = False
                while time.time() < end_time:
                    if self.state() == types.VmStatus.UP:
                        vm_resume = True
                        break
                    time.sleep(1)
                if not vm_resume:
                    raise WaitVMStateTimeoutError("RESUME", self.state())
            else:
                logging.debug('VM already up')
        except Exception as e:
            logging.error('Failed to resume VM:\n%s' % str(e))

    def shutdown(self, gracefully=True, timeout=300):
        """
        Shut down a running VM.
        """
        end_time = time.time() + timeout
        if self.is_alive():
            logging.info('Shutdown VM %s' % self.name)
            if gracefully:
                self.instance.shutdown()
            else:
                self.instance.stop()
            logging.info('Waiting for VM to reach <Down> status')
            vm_down = False
            while time.time() < end_time:
                if self.is_dead():
                    vm_down = True
                    break
                time.sleep(1)
            if not vm_down:
                raise WaitVMStateTimeoutError("DOWN", self.state())
        else:
            logging.debug('VM already down')

    def delete(self, timeout=300):
        """
        Delete a VM.
        """
        end_time = time.time() + timeout
        if self.name in self.list():
            logging.info('Delete VM %s' % self.name)
            self.instance.remove()
            logging.info('Waiting for VM to be <Deleted>')
            vm_delete = False
            while time.time() < end_time:
                if self.name not in self.list():
                    vm_delete = True
                    break
                time.sleep(1)
            if not vm_delete:
                raise WaitVMStateTimeoutError("DELETE", self.state())
            logging.info('VM was removed successfully')
        else:
            logging.debug('VM not exist')

    def destroy(self, gracefully=False):
        """
        Destroy a VM.
        """
        if not self.connection.system_service().vms_service().list():
            return
        self.shutdown(gracefully)

    def delete_from_export_domain(self, export_name):
        """
        Remove a VM from specified export domain.

        :param export_name: export domain name.
        """
        vm = self.lookup_by_storagedomains(export_name)
        try:
            sds_service = self.connection.system_service().storage_domains_service()
            export_sd = sds_service.list(search=export_name)[0]
            logging.info('Remove VM %s from export storage' % self.name)
            export_vms_service = sds_service.storage_domain_service(export_sd.id).vms_service()
            export_vms_service.vm_service(vm.id).remove()
        except Exception as e:
            logging.error('Failed to remove VM:\n%s' % str(e))

    def import_from_export_domain(self, export_name, storage_name,
                                  cluster_name, timeout=300):
        """
        Import a VM from export domain to data domain.

        :param export_name: Export domain name.
        :param storage_name: Storage domain name.
        :param cluster_name: Cluster name.
        :param timeout: timeout value
        """
        begin_time = time.time()
        end_time = time.time() + timeout
        vm = self.lookup_by_storagedomains(export_name)
        sds_service = self.connection.system_service().storage_domains_service()
        export_sd = sds_service.list(search=export_name)[0]
        storage_domains = sds_service.list(search=storage_name)[0]
        clusters_service = self.connection.system_service().clusters_service()
        cluster = clusters_service.list(search=cluster_name)[0]
        logging.info('Import VM %s' % self.name)
        export_vms_service = sds_service.storage_domain_service(export_sd.id).vms_service()
        export_vms_service.vm_service(vm.id).import_(
            storage_domain=types.StorageDomain(id=storage_domains.id),
            cluster=types.Cluster(id=cluster.id),
            vm=types.Vm(id=vm.id),
            exclusive=False
        )
        logging.info('Waiting for VM to reach <Down> status')
        vm_down = False
        while time.time() < end_time:
            if self.name in self.list():
                if self.is_dead():
                    vm_down = True
                    break
            time.sleep(1)
        if not vm_down:
            raise WaitVMStateTimeoutError("DOWN", self.state())
        logging.info('Import %s successfully(time lapse %ds)',
                     self.name, time.time() - begin_time)

    def export_from_export_domain(self, export_name, timeout=300):
        """
        Export a VM from storage domain to export domain.

        :param export_name: Export domain name.
        """
        end_time = time.time() + timeout
        storage_domains = self.connection.storagedomains.get(export_name)
        logging.info('Export VM %s' % self.name)
        self.instance.export(types.Action(storage_domain=storage_domains))
        logging.info('Waiting for VM to reach <Down> status')
        vm_down = False
        while time.time() < end_time:
            if self.is_dead():
                vm_down = True
                break
            time.sleep(1)
        if not vm_down:
            raise WaitVMStateTimeoutError("DOWN", self.state())
        logging.info('Export %s successfully', self.name)

    def snapshot(self, snapshot_name='my_snapshot', timeout=300):
        """
        Create a snapshot to VM.

        :param snapshot_name: 'my_snapshot' is default snapshot name.
        :param timeout: Time out
        """
        end_time = time.time() + timeout
        snap_params = types.Snapshot(description=snapshot_name,
                                     vm=self.instance)
        logging.info('Creating a snapshot %s for VM %s'
                     % (snapshot_name, self.name))
        self.instance.snapshots.add(snap_params)
        logging.info('Waiting for snapshot creation to finish')
        vm_snapsnop = False
        while time.time() < end_time:
            if self.state() != 'image_locked':
                vm_snapsnop = True
                break
            time.sleep(1)
        if not vm_snapsnop:
            raise WaitVMStateTimeoutError("SNAPSHOT", self.state())
        logging.info('Snapshot was created successfully')

    def create_template(self, cluster_name, template_name='my_template', timeout=300):
        """
        Create a template from VM.

        :param cluster_name: cluster name.
        :param template_name: 'my_template' is default template name.
        :param timeout: Time out
        """
        end_time = time.time() + timeout
        cluster = self.connection.clusters.get(cluster_name)

        tmpl_params = types.Template(name=template_name,
                                     vm=self.instance,
                                     cluster=cluster)
        try:
            logging.info('Creating a template %s from VM %s'
                         % (template_name, self.name))
            self.connection.templates.add(tmpl_params)
            logging.info('Waiting for VM to reach <Down> status')
            vm_down = False
            while time.time() < end_time:
                if self.is_dead():
                    vm_down = True
                    break
                time.sleep(1)
            if not vm_down:
                raise WaitVMStateTimeoutError("DOWN", self.state())
        except Exception as e:
            logging.error('Failed to create a template from VM:\n%s' % str(e))

    def add(self, memory, disk_size, cluster_name, storage_name,
            nic_name='eth0', network_interface='virtio',
            network_name='ovirtmgmt', disk_interface='virtio',
            disk_format='raw', template_name='Blank', timeout=300):
        """
        Create VM with one NIC and one Disk.

        :param memory: VM's memory size such as 1024*1024*1024=1GB.
        :param disk_size: VM's disk size such as 512*1024=512MB.
        :param nic_name: VM's NICs name such as 'eth0'.
        :param network_interface: VM's network interface such as 'virtio'.
        :param network_name: network such as ovirtmgmt for ovirt, rhevm for rhel.
        :param disk_format: VM's disk format such as 'raw' or 'cow'.
        :param disk_interface: VM's disk interface such as 'virtio'.
        :param cluster_name: cluster name.
        :param storage_name: storage domain name.
        :param template_name: VM's template name, default is 'Blank'.
        :param timeout: Time out
        """
        end_time = time.time() + timeout
        # network name is ovirtmgmt for ovirt, rhevm for rhel.
        vm_params = types.VM(name=self.name, memory=memory,
                             cluster=self.connection.clusters.get(cluster_name),
                             template=self.connection.templates.get(template_name))

        storage = self.connection.storagedomains.get(storage_name)

        storage_params = types.StorageDomains(storage_domain=[storage])

        nic_params = types.NIC(name=nic_name,
                               network=types.Network(name=network_name),
                               interface=network_interface)

        disk_params = types.Disk(storage_domains=storage_params,
                                 size=disk_size,
                                 type_='system',
                                 status=None,
                                 interface=disk_interface,
                                 format=disk_format,
                                 sparse=True,
                                 bootable=True)

        try:
            logging.info('Creating a VM %s' % self.name)
            self.connection.vms.add(vm_params)

            logging.info('NIC is added to VM %s' % self.name)
            self.instance.nics.add(nic_params)

            logging.info('Disk is added to VM %s' % self.name)
            self.instance.disks.add(disk_params)

            logging.info('Waiting for VM to reach <Down> status')
            vm_down = False
            while time.time() < end_time:
                if self.is_dead():
                    vm_down = True
                    break
                time.sleep(1)
            if not vm_down:
                raise WaitVMStateTimeoutError("DOWN", self.state())
        except Exception as e:
            logging.error('Failed to create VM with disk and NIC\n%s' % str(e))

    def add_vm_from_template(self, cluster_name, template_name='Blank',
                             new_name='my_new_vm', timeout=300):
        """
        Create a VM from template.

        :param cluster_name: cluster name.
        :param template_name: default template is 'Blank'.
        :param new_name: 'my_new_vm' is a default new VM's name.
        :param timeout: Time out
        """
        end_time = time.time() + timeout
        vm_params = types.VM(name=new_name,
                             cluster=self.connection.clusters.get(cluster_name),
                             template=self.connection.templates.get(template_name))
        try:
            logging.info('Creating a VM %s from template %s'
                         % (new_name, template_name))
            self.connection.vms.add(vm_params)
            logging.info('Waiting for VM to reach <Down> status')
            vm_down = False
            while time.time() < end_time:
                if self.is_dead():
                    vm_down = True
                    break
                time.sleep(1)
            if not vm_down:
                raise WaitVMStateTimeoutError("DOWN", self.state())
            logging.info('VM was created from template successfully')
        except Exception as e:
            logging.error('Failed to create VM from template:\n%s' % str(e))

    def get_address(self, index=0, *args):
        """
        Return the address of the guest through ovirt node tcpdump cache.

        :param index: Name or index of the NIC whose address is requested.
        :return: IP address of NIC.
        :raise VMIPAddressMissingError: If no IP address is found for the the
                NIC's MAC address
        """
        nic = self.virtnet[index]
        if nic.nettype == 'bridge':
            mac = self.get_mac_address()
            for mac_i in mac:
                ip = self.address_cache.get(mac_i)
                if ip:
                    return ip
            # TODO: Verify MAC-IP address mapping on remote ovirt node
            raise virt_vm.VMIPAddressMissingError(mac)
        else:
            raise ValueError("Ovirt only support bridge nettype now.")


class DataCenterManager(object):

    """
    This class handles all basic datacenter operations.
    """

    def __init__(self, params):
        self.name = params.get("dc_name", "")
        self.params = params
        (self.connection, self.version) = connect(params)
        self.dcs_service = self.connection.system_service().data_centers_service()

        if self.name:
            dc_search_result = self.dcs_service.list(search='name=%s' % self.name)
            if dc_search_result:
                self.instance = dc_search_result[0]

    def list(self):
        """
        List all of datacenters.
        """
        dc_list = []
        try:
            logging.info('List Data centers')
            dcs = self.dcs_service.list(search='name=%s' % self.name)
            for i in range(len(dcs)):
                dc_list.append(dcs[i].name)
            return dc_list
        except Exception as e:
            logging.error('Failed to get data centers:\n%s' % str(e))

    def add(self, storage_type):
        """
        Add a new data center.
        """
        if not self.name:
            self.name = "my_datacenter"
        try:
            logging.info('Creating a %s type datacenter %s'
                         % (storage_type, self.name))
            if self.dcs_service.add(types.DataCenter(name=self.name, storage_type=storage_type, version=self.version)):
                logging.info('Data center was created successfully')
        except Exception as e:
            logging.error('Failed to create data center:\n%s' % str(e))


class ClusterManager(object):

    """
    This class handles all basic cluster operations.
    """

    def __init__(self, params):
        self.name = params.get("cluster_name", "")
        self.params = params
        (self.connection, self.version) = connect(params)
        self.clusters_service = self.connection.system_service().clusters_service()

        if self.name:
            self.instance = self.clusters_service.list(search='name=%s' % self.name)

    def list(self):
        """
        List all of clusters.
        """
        cluster_list = []
        try:
            logging.info('List clusters')
            clusters = self.clusters_service.list()
            for i in range(len(clusters)):
                cluster_list.append(clusters[i].name)
            return cluster_list
        except Exception as e:
            logging.error('Failed to get clusters:\n%s' % str(e))

    def add(self, dc_name, cpu_type='Intel Nehalem Family'):
        """
        Add a new cluster into data center.
        """
        if not self.name:
            self.name = "my_cluster"

        dc = self.connection.system_service().data_centers_service().list(search='name=%s' % dc_name)[0]
        try:
            logging.info('Creating a cluster %s in datacenter %s'
                         % (self.name, dc_name))
            if self.clusters_service.add(types.Cluster(name=self.name,
                                                       cpu=types.CPU(id=cpu_type),
                                                       data_center=dc,
                                                       version=self.version)):
                logging.info('Cluster was created successfully')
        except Exception as e:
            logging.error('Failed to create cluster:\n%s' % str(e))


class HostManager(object):

    """
    This class handles all basic host operations.
    """

    def __init__(self, params):
        self.name = params.get("hostname", "")
        self.params = params
        (self.connection, self.version) = connect(params)
        self.hosts_service = self.connection.system_service().hosts_service()

        if self.name:
            self.instance = self.hosts_service.list(search='name=%s' % self.name)

    def list(self):
        """
        List all of hosts.
        """
        host_list = []
        try:
            logging.info('List hosts')
            hosts = self.hosts_service.list()
            for i in range(len(hosts)):
                host_list.append(hosts[i].name)
            return host_list
        except Exception as e:
            logging.error('Failed to get hosts:\n%s' % str(e))

    def state(self):
        """
        Return host state.
        """
        try:
            return self.instance.status.state
        except Exception as e:
            logging.error('Failed to get %s status:\n%s' % (self.name, str(e)))

    def add(self, host_address, host_password, cluster_name, timeout=300):
        """
        Register a host into specified cluster.
        """
        end_time = time.time() + timeout
        if not self.name:
            self.name = 'my_host'

        clusters = self.connection.system_service().clusters_service().list(
            search='name=%s' % cluster_name)[0]
        host_params = types.Host(name=self.name, address=host_address,
                                 cluster=clusters, root_password=host_password)
        try:
            logging.info('Registing a host %s into cluster %s'
                         % (self.name, cluster_name))
            if self.hosts_service.add(host_params):
                logging.info('Waiting for host to reach the <Up> status ...')
                host_up = False
                while time.time() < end_time:
                    if self.state() == types.VmStatus.UP:
                        host_up = True
                        break
                    time.sleep(1)
                if not host_up:
                    raise WaitHostStateTimeoutError("UP", self.state())
                logging.info('Host was installed successfully')
        except Exception as e:
            logging.error('Failed to install host:\n%s' % str(e))

    def get_address(self):
        """
        Return host IP address.
        """
        try:
            logging.info('Get host %s IP' % self.name)
            return self.instance.get_address()
        except Exception as e:
            logging.error('Failed to get host %s IP address:\n%s' %
                          (self.name, str(e)))


class StorageDomainManager(object):

    """
    This class handles all basic storage domain operations.
    """

    def __init__(self, params):
        self.name = params.get("storage_name", "")
        self.params = params
        (self.connection, self.version) = connect(params)
        self.sds_service = self.connection.system_service().storage_domains_service()

        if self.name:
            self.instance = self.sds_service.list(search='name=%s' % self.name)

    def list(self):
        """
        List all of storagedomains.
        """
        storage_list = []
        try:
            logging.info('List storage domains')
            storages = self.sds_service.list()
            for i in range(len(storages)):
                storage_list.append(storages[i].name)
            return storage_list
        except Exception as e:
            logging.error('Failed to get storage domains:\n%s' % str(e))

    def attach_iso_export_domain_into_datacenter(self, address, path,
                                                 dc_name, host_name,
                                                 domain_type,
                                                 storage_type='nfs',
                                                 name='my_iso'):
        """
        Attach ISO/export domain into data center.

        :param name: ISO or Export name.
        :param host_name: host name.
        :param dc_name: data center name.
        :param path: ISO/export domain path.
        :param address: ISO/export domain address.
        :param domain_type: storage domain type, it may be 'iso' or 'export'.
        :param storage_type: storage type, it may be 'nfs', 'iscsi', or 'fc'.
        """
        dc = self.api.datacenters.get(dc_name)
        host = self.api.hosts.get(host_name)
        storage_params = types.Storage(type_=storage_type,
                                       address=address,
                                       path=path)

        storage_domain__params = types.StorageDomain(name=name,
                                                     data_center=dc,
                                                     type_=domain_type,
                                                     host=host,
                                                     storage=storage_params)

        try:
            logging.info('Create/import ISO storage domain %s' % name)
            if self.api.storagedomains.add(storage_domain__params):
                logging.info('%s domain was created/imported successfully'
                             % domain_type)

            logging.info('Attach ISO storage domain %s' % name)
            if self.api.datacenters.get(dc_name).storagedomains.add(
                    self.api.storagedomains.get(name)):
                logging.info('%s domain was attached successfully'
                             % domain_type)

            logging.info('Activate ISO storage domain %s' % name)
            if self.api.datacenters.get(dc_name).storagedomains.get(
                    name).activate():
                logging.info('%s domain was activated successfully'
                             % domain_type)
        except Exception as e:
            logging.error('Failed to add %s domain:\n%s'
                          % (domain_type, str(e)))



import logging

from six.moves import xrange

from . import env_process
from . import error_context
from . import qemu_virtio_port


class VirtioPortTest(object):

    def __init__(self, test, env, params):
        self.test = test
        self.env = env
        self.params = params

    @error_context.context_aware
    def get_vm_with_ports(self, no_consoles=0, no_serialports=0, spread=None,
                          quiet=False, strict=False):
        """
        Checks whether existing 'main_vm' fits the requirements, modifies
        it if needed and returns the VM object.

        :param no_console: Number of desired virtconsoles.
        :param no_serialport: Number of desired virtserialports.
        :param spread: Spread consoles across multiple virtio-serial-pcis.
        :param quiet: Notify user about VM recreation.
        :param strict: Whether no_consoles have to match or just exceed.
        :return: vm object matching the requirements.
        """
        params = self.params.copy()
        main_vm = self.params['main_vm']
        # check the number of running VM's consoles
        vm = self.env.get_vm(main_vm)

        if not vm:
            _no_serialports = -1
            _no_consoles = -1
        else:
            _no_serialports = 0
            _no_consoles = 0
            for port in vm.virtio_ports:
                if isinstance(port, qemu_virtio_port.VirtioSerial):
                    _no_serialports += 1
                else:
                    _no_consoles += 1
        _spread = int(params.get('virtio_port_spread', 2))
        if spread is None:
            spread = _spread
        if strict:
            if (_no_serialports != no_serialports or
                    _no_consoles != no_consoles):
                _no_serialports = -1
                _no_consoles = -1
        # If not enough ports, modify params and recreate VM
        if (_no_serialports < no_serialports or
                _no_consoles < no_consoles or
                spread != _spread):
            if not quiet:
                out = "tests requirements are different from cfg: "
                if _no_serialports < no_serialports:
                    out += "serial_ports(%d), " % no_serialports
                if _no_consoles < no_consoles:
                    out += "consoles(%d), " % no_consoles
                if spread != _spread:
                    out += "spread(%s), " % spread
                logging.warning(out[:-2] + ". Modify config to speedup tests.")

            params['serials'] = params.objects('serials')[0]
            if spread:
                params['virtio_port_spread'] = spread
            else:
                params['virtio_port_spread'] = 0

            for i in xrange(max(no_consoles, _no_consoles)):
                name = "console-%d" % i
                params['serials'] += " %s" % name
                params['serial_type_%s' % name] = "virtconsole"

            for i in xrange(max(no_serialports, _no_serialports)):
                name = "serialport-%d" % i
                params['serials'] += " %s" % name
                params['serial_type_%s' % name] = "virtserialport"

            if quiet:
                logging.debug("Recreating VM with more virtio ports.")
            else:
                logging.warning("Recreating VM with more virtio ports.")
            env_process.preprocess_vm(self.test, params, self.env, main_vm)
            vm = self.env.get_vm(main_vm)

        vm.verify_kernel_crash()
        return vm

    @error_context.context_aware
    def get_vm_with_worker(self, no_consoles=0, no_serialports=0, spread=None,
                           quiet=False):
        """
        Checks whether existing 'main_vm' fits the requirements, modifies
        it if needed and returns the VM object and guest_worker.

        :param no_console: Number of desired virtconsoles.
        :param no_serialport: Number of desired virtserialports.
        :param spread: Spread consoles across multiple virtio-serial-pcis.
        :param quiet: Notify user about VM recreation.
        :param strict: Whether no_consoles have to match or just exceed.
        :return: tuple (vm object matching the requirements,
                        initialized GuestWorker of the vm)
        """
        vm = self.get_vm_with_ports(no_consoles, no_serialports, spread, quiet)
        guest_worker = qemu_virtio_port.GuestWorker(vm)
        return vm, guest_worker

    @error_context.context_aware
    def get_vm_with_single_port(self, port_type='serialport'):
        """
        Wrapper which returns vm, guest_worker and virtio_ports with at lest
        one port of the type specified by fction parameter.

        :param port_type: type of the desired virtio port.
        :return: tuple (vm object with at least 1 port of the port_type,
                        initialized GuestWorker of the vm,
                        list of virtio_ports of the port_type type)
        """
        if port_type == 'serialport':
            vm, guest_worker = self.get_vm_with_worker(no_serialports=1)
            virtio_ports = self.get_virtio_ports(vm)[1][0]
        else:
            vm, guest_worker = self.get_vm_with_worker(no_consoles=1)
            virtio_ports = self.get_virtio_ports(vm)[0][0]
        return vm, guest_worker, virtio_ports

    @error_context.context_aware
    def get_virtio_ports(self, vm):
        """
        Returns separated virtconsoles and virtserialports

        :param vm: VM object
        :return: tuple (all virtconsoles, all virtserialports)
        """
        consoles = []
        serialports = []
        for port in vm.virtio_ports:
            if isinstance(port, qemu_virtio_port.VirtioSerial):
                serialports.append(port)
            else:
                consoles.append(port)
        return (consoles, serialports)

    @staticmethod
    @error_context.context_aware
    def cleanup(vm=None, guest_worker=None):
        """
        Cleanup function.

        :param vm: VM whose ports should be cleaned
        :param guest_worker: guest_worker which should be cleaned/exited
        """
        error_context.context("Cleaning virtio_ports on guest.")
        if guest_worker:
            guest_worker.cleanup()
        error_context.context("Cleaning virtio_ports on host.")
        if vm:
            for port in vm.virtio_ports:
                port.clean_port()
                port.close()
                port.mark_as_clean()



import os
import logging
import sys

from autotest.client import test
from autotest.client import utils
from autotest.client.shared import git, error, software_manager


class kernelinstall(test.test):
    version = 1
    sm = software_manager.SoftwareManager()

    def _kernel_install_rpm(self, rpm_file, kernel_deps_rpms=None,
                            need_reboot=True):
        """
        Install kernel rpm package.
        The rpm packages should be a url or put in this test's
        directory (client/test/kernelinstall)
        """
        if kernel_deps_rpms:
            logging.info("Installing kernel dependencies.")
            if isinstance(kernel_deps_rpms, list):
                kernel_deps_rpms = " ".join(kernel_deps_rpms)
            self.sm.install(kernel_deps_rpms)

        dst = os.path.join("/tmp", os.path.basename(rpm_file))
        knl = utils.get_file(rpm_file, dst)
        kernel = self.job.kernel(knl)
        logging.info("Installing kernel %s", rpm_file)
        kernel.install(install_vmlinux=False)

        if need_reboot:
            kernel.boot()
        else:
            kernel.add_to_bootloader()

    def _kernel_install_koji(self, kernel_koji_spec, kernel_deps_koji_spec,
                             need_reboot=True):
        # Using hardcoded package names (the names are not expected to change)
        # we avoid lookup errors due to SSL problems, so let's go with that.
        for koji_package in ['koji', 'brewkoji']:
            if not self.sm.check_installed(koji_package):
                logging.debug("%s missing - trying to install", koji_package)
                self.sm.install(koji_package)

        sys.path.append(self.bindir)
        try:
            from staging import utils_koji
        except ImportError:
            from autotest.client.shared import utils_koji
        # First, download packages via koji/brew
        c = utils_koji.KojiClient()

        deps_rpms = []
        k_dep = utils_koji.KojiPkgSpec(text=kernel_deps_koji_spec)
        logging.info('Fetching kernel dependencies: %s', kernel_deps_koji_spec)
        c.get_pkgs(k_dep, self.bindir)
        rpm_file_name_list = c.get_pkg_rpm_file_names(k_dep)
        if len(rpm_file_name_list) == 0:
            raise error.TestError("No packages on brew/koji match spec %s" %
                                  kernel_deps_koji_spec)
        dep_rpm_basename = rpm_file_name_list[0]
        deps_rpms.append(os.path.join(self.bindir, dep_rpm_basename))

        k = utils_koji.KojiPkgSpec(text=kernel_koji_spec)
        logging.info('Fetching kernel: %s', kernel_koji_spec)
        c.get_pkgs(k, self.bindir)
        rpm_file_name_list = c.get_pkg_rpm_file_names(k)
        if len(rpm_file_name_list) == 0:
            raise error.TestError("No packages on brew/koji match spec %s" %
                                  kernel_koji_spec)

        kernel_rpm_basename = rpm_file_name_list[0]
        kernel_rpm_path = os.path.join(self.bindir, kernel_rpm_basename)

        # Then install kernel rpm packages.
        self._kernel_install_rpm(kernel_rpm_path, deps_rpms, need_reboot)

    def _kernel_install_src(self, base_tree, config=None, config_list=None,
                            patch_list=None, need_reboot=True):
        if not utils.is_url(base_tree):
            base_tree = os.path.join(self.bindir, base_tree)
        if not utils.is_url(config):
            config = os.path.join(self.bindir, config)
        kernel = self.job.kernel(base_tree, self.outputdir)
        if patch_list:
            patches = []
            for p in patch_list.split():
                # Make sure all the patches are in local.
                if not utils.is_url(p):
                    continue
                dst = os.path.join(self.bindir, os.path.basename(p))
                local_patch = utils.get_file(p, dst)
                patches.append(local_patch)
            kernel.patch(*patches)
        if not os.path.isfile(config):
            config = None
        if not config and not config_list:
            kernel.config()
        else:
            kernel.config(config, config_list)
        kernel.build()
        kernel.install()

        if need_reboot:
            kernel.boot()
        else:
            kernel.add_to_bootloader()

    def _kernel_install_git(self, repo, config, repo_base=None,
                            branch="master", commit=None, config_list=None,
                            patch_list=None, need_reboot=True):
        repodir = os.path.join("/tmp", 'kernel_src')
        repodir = git.get_repo(uri=repo, branch=branch,
                               destination_dir=repodir,
                               commit=commit, base_uri=repo_base)
        self._kernel_install_src(repodir, config, config_list, patch_list,
                                 need_reboot)

    def execute(self, install_type="koji", params=None):
        need_reboot = params.get("need_reboot") == "yes"

        logging.info("Chose to install kernel through '%s', proceeding",
                     install_type)

        if install_type == "rpm":
            rpm_url = params.get("kernel_rpm_path")
            kernel_deps_rpms = params.get("kernel_deps_rpms", None)

            self._kernel_install_rpm(rpm_url, kernel_deps_rpms, need_reboot)
        elif install_type in ["koji", "brew"]:

            kernel_koji_spec = params.get("kernel_koji_spec")
            kernel_deps_koji_spec = params.get("kernel_deps_koji_spec")

            self._kernel_install_koji(kernel_koji_spec, kernel_deps_koji_spec,
                                      need_reboot)

        elif install_type == "git":
            repo = params.get('kernel_git_repo')
            repo_base = params.get('kernel_git_repo_base', None)
            branch = params.get('kernel_git_branch', "master")
            commit = params.get('kernel_git_commit', None)
            patch_list = params.get("kernel_patch_list", None)
            config = params.get('kernel_config')
            config_list = params.get("kernel_config_list", None)

            self._kernel_install_git(repo, config, repo_base, branch, commit,
                                     config_list, patch_list, need_reboot)
        elif install_type == "tar":
            src_pkg = params.get("kernel_src_pkg")
            config = params.get('kernel_config')
            patch_list = params.get("kernel_patch_list", None)

            self._kernel_install_src(src_pkg, config, None, patch_list,
                                     need_reboot)
        else:
            logging.error("Could not find '%s' method, "
                          "keep the current kernel.", install_type)



#!/usr/bin/python

"""
Cartesian configuration format file parser.

Filter syntax:

* ``,`` means ``OR``
* ``..`` means ``AND``
* ``.`` means ``IMMEDIATELY-FOLLOWED-BY``
* ``(xx=yy)`` where ``xx=VARIANT_NAME`` and ``yy=VARIANT_VALUE``

Example:

::

     qcow2..(guest_os=Fedora).14, RHEL.6..raw..boot, smp2..qcow2..migrate..ide

means match all dicts whose names have:

::

    (qcow2 AND ((guest_os=Fedora) IMMEDIATELY-FOLLOWED-BY 14)) OR
    ((RHEL IMMEDIATELY-FOLLOWED-BY 6) AND raw AND boot) OR
    (smp2 AND qcow2 AND migrate AND ide)

Note:

* ``qcow2..Fedora.14`` is equivalent to ``Fedora.14..qcow2``.
* ``qcow2..Fedora.14`` is not equivalent to ``qcow2..14.Fedora``.
* ``ide, scsi`` is equivalent to ``scsi, ide``.

Filters can be used in 3 ways:

::

    only <filter>
    no <filter>
    <filter>:

The last one starts a conditional block.

Formal definition: Regexp come from `python <http://docs.python.org/2/library/re.html>`__.
They're not deterministic, but more readable for people. Spaces between
terminals and nonterminals are only for better reading of definitions.

The base of the definitions come verbatim as follows:


::

    E = {\\n, #, :, "-", =, +=, <=, ~=, ?=, ?+=, ?<=, !, < , del, @, variants, include, only, no, name, value}

    N = {S, DEL, FILTER, FILTER_NAME, FILTER_GROUP, PN_FILTER_GROUP, STAT, VARIANT, VAR-TYPE, VAR-NAME, VAR-NAME-F, VAR, COMMENT, TEXT, DEPS, DEPS-NAME-F, META-DATA, IDENTIFIER}``


    I = I^n | n in N              // indentation from start of line
                                  // where n is indentation length.
    I = I^n+x | n,x in N          // indentation with shift

    start symbol = S
    end symbol = eps

    S -> I^0+x STATV | eps

    I^n    STATV
    I^n    STATV

    I^n STATV -> I^n STATV \\n I^n STATV | I^n STAT | I^n variants VARIANT
    I^n STAT -> I^n STAT \\n I^n STAT | I^n COMMENT | I^n include INC
    I^n STAT -> I^n del DEL | I^n FILTER

    DEL -> name \\n

    I^n STAT -> I^n name = VALUE | I^n name += VALUE | I^n name <= VALUE | I^n name ~= VALUE
    I^n STAT -> I^n name ?= VALUE | I^n name ?+= VALUE | I^n name ?<= VALUE

    VALUE -> TEXT \\n | 'TEXT' \\n | "TEXT" \\n

    COMMENT_BLOCK -> #TEXT | //TEXT
    COMMENT ->  COMMENT_BLOCK\\n
    COMMENT ->  COMMENT_BLOCK\\n

    TEXT = [^\\n] TEXT            //python format regexp

    I^n    variants VAR #comments:             add possibility for comment
    I^n+x       VAR-NAME: DEPS
    I^n+x+x2        STATV
    I^n         VAR-NAME:

    IDENTIFIER -> [A-Za-z0-9][A-Za-z0-9_-]*

    VARIANT -> VAR COMMENT_BLOCK\\n I^n+x VAR-NAME
    VAR -> VAR-TYPE: | VAR-TYPE META-DATA: | :         // Named | unnamed variant

    VAR-TYPE -> IDENTIFIER

    variants _name_ [xxx] [zzz=yyy] [uuu]:

    META-DATA -> [IDENTIFIER] | [IDENTIFIER=TEXT] | META-DATA META-DATA

    I^n VAR-NAME -> I^n VAR-NAME \\n I^n VAR-NAME | I^n VAR-NAME-N \\n I^n+x STATV
    VAR-NAME-N -> - @VAR-NAME-F: DEPS | - VAR-NAME-F: DEPS
    VAR-NAME-F -> [a-zA-Z0-9\\._-]+                  // Python regexp

    DEPS -> DEPS-NAME-F | DEPS-NAME-F,DEPS
    DEPS-NAME-F -> [a-zA-Z0-9\\._- ]+                // Python regexp

    INC -> name \\n


    FILTER_GROUP: STAT
        STAT

    I^n STAT -> I^n PN_FILTER_GROUP | I^n ! PN_FILTER_GROUP

    PN_FILTER_GROUP -> FILTER_GROUP: \\n I^n+x STAT
    PN_FILTER_GROUP -> FILTER_GROUP: STAT \\n I^n+x STAT

    only FILTER_GROUP
    no FILTER_GROUP

    FILTER -> only FILTER_GROUP \\n | no FILTER_GROUP \\n

    FILTER_GROUP -> FILTER_NAME
    FILTER_GROUP -> FILTER_GROUP..FILTER_GROUP
    FILTER_GROUP -> FILTER_GROUP,FILTER_GROUP

    FILTER_NAME -> FILTER_NAME.FILTER_NAME
    FILTER_NAME -> VAR-NAME-F | (VAR-NAME-F=VAR-NAME-F)

:copyright: Red Hat 2008-2013
"""

import os
import collections
import optparse
import logging
import re
import sys

_reserved_keys = set(("name", "shortname", "dep", "_short_name_map_file", "_name_map_file"))

num_failed_cases = 5


class ParserError(Exception):

    def __init__(self, msg, line=None, filename=None, linenum=None):
        Exception.__init__(self)
        self.msg = msg
        self.line = line
        self.filename = filename
        self.linenum = linenum

    def __str__(self):
        if self.line:
            return "%s: %r (%s:%s)" % (self.msg, self.line,
                                       self.filename, self.linenum)
        else:
            return "%s (%s:%s)" % (self.msg, self.filename, self.linenum)


class LexerError(ParserError):
    pass


class MissingIncludeError(Exception):

    def __init__(self, line, filename, linenum):
        Exception.__init__(self)
        self.line = line
        self.filename = filename
        self.linenum = linenum

    def __str__(self):
        return ("%r (%s:%s): file does not exist or it's not a regular "
                "file" % (self.line, self.filename, self.linenum))


if sys.version_info[0] == 2 and sys.version_info[1] < 6:
    def enum(iterator, start_pos=0):
        for i in iterator:
            yield start_pos, i
            start_pos += 1
else:
    enum = enumerate


def _match_adjacent(block, ctx, ctx_set):
    """
    It try to match as many blocks as possible from context.

    :return: Count of matched blocks.
    """
    if block[0] not in ctx_set:
        return 0
    if len(block) == 1:
        return 1                          # First match and length is 1.
    if block[1] not in ctx_set:
        return int(ctx[-1] == block[0])   # Check match with last from ctx.
    k = 0
    i = ctx.index(block[0])
    while i < len(ctx):                   # Try to  match all of blocks.
        if k > 0 and ctx[i] != block[k]:  # Block not match
            i -= k - 1
            k = 0                         # Start from first block in next ctx.
        if ctx[i] == block[k]:
            k += 1
            if k >= len(block):           # match all of blocks
                break
            if block[k] not in ctx_set:   # block in not in whole ctx.
                break
        i += 1
    return k


def _might_match_adjacent(block, ctx, ctx_set, descendant_labels):
    matched = _match_adjacent(block, ctx, ctx_set)
    for elem in block[matched:]:        # Try to find rest of blocks in subtree
        if elem not in descendant_labels:
            # print "Can't match %s, ctx %s" % (block, ctx)
            return False
    return True


# Filter must inherit from object (otherwise type() won't work)
class Filter(object):
    __slots__ = ["filter"]

    def __init__(self, lfilter):
        self.filter = lfilter
        # print self.filter

    def match(self, ctx, ctx_set):
        for word in self.filter:  # Go through ,
            for block in word:    # Go through ..
                if _match_adjacent(block, ctx, ctx_set) != len(block):
                    break
            else:
                # print "Filter pass: %s ctx: %s" % (self.filter, ctx)
                return True       # All match
        return False

    def might_match(self, ctx, ctx_set, descendant_labels):
        # There is some posibility to match in children blocks.
        for word in self.filter:
            for block in word:
                if not _might_match_adjacent(block, ctx, ctx_set,
                                             descendant_labels):
                    break
            else:
                return True
        # print "Filter not pass: %s ctx: %s" % (self.filter, ctx)
        return False


class NoOnlyFilter(Filter):
    __slots__ = ("line")

    def __init__(self, lfilter, line):
        super(NoOnlyFilter, self).__init__(lfilter)
        self.line = line

    def __eq__(self, o):
        if isinstance(o, self.__class__):
            if self.filter == o.filter:
                return True

        return False


class OnlyFilter(NoOnlyFilter):
    # pylint: disable=W0613

    def is_irrelevant(self, ctx, ctx_set, descendant_labels):
        # Matched in this tree.
        return self.match(ctx, ctx_set)

    def requires_action(self, ctx, ctx_set, descendant_labels):
        # Impossible to match in this tree.
        return not self.might_match(ctx, ctx_set, descendant_labels)

    def might_pass(self, failed_ctx, failed_ctx_set, ctx, ctx_set,
                   descendant_labels):
        for word in self.filter:
            for block in word:
                if (_match_adjacent(block, ctx, ctx_set) >
                        _match_adjacent(block, failed_ctx, failed_ctx_set)):
                    return self.might_match(ctx, ctx_set, descendant_labels)
        return False

    def __str__(self):
        return "Only %s" % (self.filter)

    def __repr__(self):
        return "Only %s" % (self.filter)


class NoFilter(NoOnlyFilter):

    def is_irrelevant(self, ctx, ctx_set, descendant_labels):
        return not self.might_match(ctx, ctx_set, descendant_labels)

    # pylint: disable=W0613
    def requires_action(self, ctx, ctx_set, descendant_labels):
        return self.match(ctx, ctx_set)

    # pylint: disable=W0613
    def might_pass(self, failed_ctx, failed_ctx_set, ctx, ctx_set,
                   descendant_labels):
        for word in self.filter:
            for block in word:
                if (_match_adjacent(block, ctx, ctx_set) <
                        _match_adjacent(block, failed_ctx, failed_ctx_set)):
                    return not self.match(ctx, ctx_set)
        return False

    def __str__(self):
        return "No %s" % (self.filter)

    def __repr__(self):
        return "No %s" % (self.filter)


class JoinFilter(NoOnlyFilter):

    def __str__(self):
        return "Join %s" % (self.filter)

    def __repr__(self):
        return "Join %s" % (self.filter)


class BlockFilter(object):
    __slots__ = ["blocked"]

    def __init__(self, blocked):
        self.blocked = blocked

    def apply_to_dict(self, d):
        pass


class Condition(NoFilter):
    __slots__ = ["content"]

    # pylint: disable=W0231
    def __init__(self, lfilter, line):
        super(Condition, self).__init__(lfilter, line)
        self.content = []

    def __str__(self):
        return "Condition %s:%s" % (self.filter, self.content)

    def __repr__(self):
        return "Condition %s:%s" % (self.filter, self.content)


class NegativeCondition(OnlyFilter):
    __slots__ = ["content"]

    # pylint: disable=W0231
    def __init__(self, lfilter, line):
        super(NegativeCondition, self).__init__(lfilter, line)
        self.content = []

    def __str__(self):
        return "NotCond %s:%s" % (self.filter, self.content)

    def __repr__(self):
        return "NotCond %s:%s" % (self.filter, self.content)


class StrReader(object):

    """
    Preprocess an input string for easy reading.
    """

    def __init__(self, s):
        """
        Initialize the reader.

        :param s: The string to parse.
        """
        self.filename = "<string>"
        self._lines = []
        self._line_index = 0
        self._stored_line = None
        for linenum, line in enumerate(s.splitlines()):
            line = line.rstrip().expandtabs()
            stripped_line = line.lstrip()
            indent = len(line) - len(stripped_line)
            if (not stripped_line or
                    stripped_line.startswith("#") or
                    stripped_line.startswith("//")):
                continue
            self._lines.append((stripped_line, indent, linenum + 1))

    def get_next_line(self, prev_indent):
        """
        Get the next line in the current block.

        :param prev_indent: The indentation level of the previous block.
        :return: (line, indent, linenum), where indent is the line's
            indentation level.  If no line is available, (None, -1, -1) is
            returned.
        """
        if self._stored_line:
            ret = self._stored_line
            self._stored_line = None
            return ret
        if self._line_index >= len(self._lines):
            return None, -1, -1
        line, indent, linenum = self._lines[self._line_index]
        if indent <= prev_indent:
            return None, indent, linenum
        self._line_index += 1
        return line, indent, linenum

    def set_next_line(self, line, indent, linenum):
        """
        Make the next call to get_next_line() return the given line instead of
        the real next line.
        """
        line = line.strip()
        if line:
            self._stored_line = line, indent, linenum


class FileReader(StrReader):

    """
    Preprocess an input file for easy reading.
    """

    def __init__(self, filename):
        """
        Initialize the reader.

        :parse filename: The name of the input file.
        """
        with open(filename) as f:
            StrReader.__init__(self, f.read())
        self.filename = filename


class Label(object):
    __slots__ = ["name", "var_name", "long_name", "hash_val", "hash_var"]

    def __init__(self, name, next_name=None):
        if next_name is None:
            self.name = name
            self.var_name = None
        else:
            self.name = next_name
            self.var_name = name

        if self.var_name is None:
            self.long_name = "%s" % (self.name)
        else:
            self.long_name = "(%s=%s)" % (self.var_name, self.name)

        self.hash_val = self.hash_name()
        self.hash_var = None
        if self.var_name:
            self.hash_var = self.hash_variant()

    def __str__(self):
        return self.long_name

    def __repr__(self):
        return self.long_name

    def __eq__(self, o):
        """
        The comparison is asymmetric due to optimization.
        """
        if o.var_name:
            if self.long_name == o.long_name:
                return True
        else:
            if self.name == o.name:
                return True
        return False

    def __ne__(self, o):
        """
        The comparison is asymmetric due to optimization.
        """
        if o.var_name:
            if self.long_name != o.long_name:
                return True
        else:
            if self.name != o.name:
                return True
        return False

    def __hash__(self):
        return self.hash_val

    def hash_name(self):
        return sum([i + 1 * ord(x) for i, x in enumerate(self.name)])

    def hash_variant(self):
        return sum([i + 1 * ord(x) for i, x in enumerate(str(self))])


class Node(object):
    __slots__ = ["var_name", "name", "filename", "dep", "content", "children",
                 "labels", "append_to_shortname", "failed_cases", "default",
                 "q_dict"]

    def __init__(self):
        self.var_name = []
        self.name = []
        self.filename = ""
        self.dep = []
        self.content = []
        self.children = []
        self.labels = set()
        self.append_to_shortname = False
        self.failed_cases = collections.deque()
        self.default = False

    def dump(self, indent, recurse=False):
        print("%s%s" % (" " * indent, self.name))
        print("%s%s" % (" " * indent, self.var_name))
        print("%s%s" % (" " * indent, self))
        print("%s%s" % (" " * indent, self.content))
        print("%s%s" % (" " * indent, self.failed_cases))
        if recurse:
            for child in self.children:
                child.dump(indent + 3, recurse)


match_substitute = re.compile("\$\{(.+?)\}")


def _substitution(value, d):
    """
    Only optimization string Template substitute is quite expensive operation.

    :param value: String where could be $string for substitution.
    :param d: Dictionary from which should be value substituted to value.

    :return: Substituted string
    """
    if "$" in value:
        start = 0
        st = ""
        try:
            match = match_substitute.search(value, start)
            while match:
                val = eval(match.group(1), None, d)
                st += value[start:match.start()] + str(val)
                start = match.end()
                match = match_substitute.search(value, start)
        except Exception:
            pass
        st += value[start:len(value)]
        return st
    else:
        return value


class Token(object):
    __slots__ = []
    identifier = ""

    def __str__(self):
        return self.identifier

    def __repr__(self):
        return "'%s'" % self.identifier

    def __ne__(self, o):
        """
        The comparison is asymmetric due to optimization.
        """
        if o.identifier != self.identifier:
            return True
        return False


class LIndent(Token):
    __slots__ = ["length"]
    identifier = "indent"

    def __init__(self, length):
        self.length = length

    def __str__(self):
        return "%s %s" % (self.identifier, self.length)

    def __repr__(self):
        return "%s %s" % (self.identifier, self.length)


class LEndL(Token):
    __slots__ = []
    identifier = "endl"


class LEndBlock(LIndent):
    __slots__ = []
    pass


class LIdentifier(str):
    __slots__ = []
    identifier = "Identifier re([A-Za-z0-9][A-Za-z0-9_-]*)"

    def __str__(self):
        return super(LIdentifier, self).__str__()

    def __repr__(self):
        return "'%s'" % self

    def checkChar(self, chars):
        for t in self:
            if not (t in chars):
                raise ParserError("Wrong char %s in %s" % (t, self))
        return self

    def checkAlpha(self):
        """
        Check if string contain only chars
        """
        if not self.isalpha():
            raise ParserError("Some of chars is not alpha in %s" % (self))
        return self

    def checkNumbers(self):
        """
        Check if string contain only chars
        """
        if not self.isdigit():
            raise ParserError("Some of chars is not digit in %s" % (self))
        return self

    def checkCharAlpha(self, chars):
        """
        Check if string contain only chars
        """
        for t in self:
            if not (t in chars or t.isalpha()):
                raise ParserError("Char %s is not alpha or one of special"
                                  "chars [%s] in %s" % (t, chars, self))
        return self

    def checkCharAlphaNum(self, chars):
        """
        Check if string contain only chars
        """
        for t in self:
            if not (t in chars or t.isalnum()):
                raise ParserError("Char %s is not alphanum or one of special"
                                  "chars [%s] in %s" % (t, chars, self))
        return self

    def checkCharNumeric(self, chars):
        """
        Check if string contain only chars
        """
        for t in self:
            if not (t in chars or t.isdigit()):
                raise ParserError("Char %s is not digit or one of special"
                                  "chars [%s] in %s" % (t, chars, self))
        return self


class LWhite(LIdentifier):
    __slots__ = []
    identifier = "WhiteSpace re(\\s)"


class LString(LIdentifier):
    __slots__ = []
    identifier = "String re(.+)"


class LColon(Token):
    __slots__ = []
    identifier = ":"


class LVariants(Token):
    __slots__ = []
    identifier = "variants"


class LDot(Token):
    __slots__ = []
    identifier = "."


class LVariant(Token):
    __slots__ = []
    identifier = "-"


class LDefault(Token):
    __slots__ = []
    identifier = "@"


class LOnly(Token):
    __slots__ = []
    identifier = "only"


class LSuffix(Token):
    __slots__ = []
    identifier = "suffix"


class LJoin(Token):
    __slots__ = []
    identifier = "join"


class LNo(Token):
    __slots__ = []
    identifier = "no"


class LCond(Token):
    __slots__ = []
    identifier = ""


class LNotCond(Token):
    __slots__ = []
    identifier = "!"


class LOr(Token):
    __slots__ = []
    identifier = ","


class LAnd(Token):
    __slots__ = []
    identifier = ".."


class LCoc(Token):
    __slots__ = []
    identifier = "."


class LComa(Token):
    __slots__ = []
    identifier = ","


class LLBracket(Token):
    __slots__ = []
    identifier = "["


class LRBracket(Token):
    __slots__ = []
    identifier = "]"


class LLRBracket(Token):
    __slots__ = []
    identifier = "("


class LRRBracket(Token):
    __slots__ = []
    identifier = ")"


class LRegExpStart(Token):
    __slots__ = []
    identifier = "${"


class LRegExpStop(Token):
    __slots__ = []
    identifier = "}"


class LInclude(Token):
    __slots__ = []
    identifier = "include"


class LOperators(Token):
    __slots__ = ["name", "value"]
    identifier = ""
    function = None

    def set_operands(self, name, value):
        # pylint: disable=W0201
        self.name = str(name)
        # pylint: disable=W0201
        self.value = str(value)
        return self


class LSet(LOperators):
    __slots__ = []
    identifier = "="

    def apply_to_dict(self, d):
        """
        :param d: Dictionary for apply value
        """
        if self.name not in _reserved_keys:
            d[self.name] = _substitution(self.value, d)


class LAppend(LOperators):
    __slots__ = []
    identifier = "+="

    def apply_to_dict(self, d):
        if self.name not in _reserved_keys:
            d[self.name] = d.get(self.name, "") + _substitution(self.value, d)


class LPrepend(LOperators):
    __slots__ = []
    identifier = "<="

    def apply_to_dict(self, d):
        if self.name not in _reserved_keys:
            d[self.name] = _substitution(self.value, d) + d.get(self.name, "")


class LLazySet(LOperators):
    __slots__ = []
    identifier = "~="

    def apply_to_dict(self, d):
        if self.name not in _reserved_keys and self.name not in d:
            d[self.name] = _substitution(self.value, d)


class LRegExpSet(LOperators):
    __slots__ = []
    identifier = "?="

    def apply_to_dict(self, d):
        exp = re.compile("%s$" % self.name)
        value = _substitution(self.value, d)
        for key in d:
            keystr = "".join(key) if isinstance(key, tuple) else key
            if key not in _reserved_keys and exp.match(keystr):
                d[key] = value


class LRegExpAppend(LOperators):
    __slots__ = []
    identifier = "?+="

    def apply_to_dict(self, d):
        exp = re.compile("%s$" % self.name)
        value = _substitution(self.value, d)
        for key in d:
            keystr = "".join(key) if isinstance(key, tuple) else key
            if key not in _reserved_keys and exp.match(keystr):
                d[key] += value


class LRegExpPrepend(LOperators):
    __slots__ = []
    identifier = "?<="

    def apply_to_dict(self, d):
        exp = re.compile("%s$" % self.name)
        value = _substitution(self.value, d)
        for key in d:
            keystr = "".join(key) if isinstance(key, tuple) else key
            if key not in _reserved_keys and exp.match(keystr):
                d[key] = value + d[key]


class LDel(LOperators):
    __slots__ = []
    identifier = "del"

    def apply_to_dict(self, d):
        exp = re.compile("%s$" % self.name)
        keys_to_del = collections.deque()
        for key in d:
            if key not in _reserved_keys and exp.match(key):
                keys_to_del.append(key)
        for key in keys_to_del:
            del d[key]


class LApplyPreDict(LOperators):
    __slots__ = []
    identifier = "apply_pre_dict"

    def set_operands(self, name, value):
        self.name = name    # pylint: disable=W0201,E0237
        self.value = value  # pylint: disable=W0201,E0237
        return self

    def apply_to_dict(self, d):
        d.update(self.value)

    def __str__(self):
        return "Apply_pre_dict: %s" % self.value

    def __repr__(self):
        return "Apply_pre_dict: %s" % self.value


class LUpdateFileMap(LOperators):
    __slots__ = ["shortname", "dest"]
    identifier = "update_file_map"

    def set_operands(self, filename, name, dest="_name_map_file"):
        # pylint: disable=W0201
        self.name = name
        # pylint: disable=W0201
        if filename == "<string>":
            self.shortname = filename
        else:
            self.shortname = os.path.basename(filename)

        self.dest = dest
        return self

    def apply_to_dict(self, d):
        dest = self.dest
        if dest not in d:
            d[dest] = {}

        if self.shortname in d[dest]:
            old_name = d[dest][self.shortname]
            d[dest][self.shortname] = "%s.%s" % (self.name, old_name)
        else:
            d[dest][self.shortname] = self.name


class Suffix(LOperators):
    __slots__ = []
    identifier = "apply_suffix"

    def __str__(self):
        return "Suffix: %s" % (self.value)

    def __repr__(self):
        return "Suffix %s" % (self.value)

    def __eq__(self, o):
        if isinstance(o, self.__class__):
            if self.value == o.value:
                return True

    def apply_to_dict(self, d):
        for key in d.copy():
            if key not in _reserved_keys:
                # Store key as a tuple: (key, suffix1, suffix2, suffix3,....)
                # This allows us to manipulate later on suffixes
                # Add suffix to the key, remove the old key
                new_key = (key if isinstance(key, tuple) else (key,)) + (self.value,)
                d[new_key] = d.pop(key)


spec_iden = "_-"
spec_oper = "+<?~"


tokens_map = {"-": LVariant,
              ".": LDot,
              ":": LColon,
              "@": LDefault,
              ",": LComa,
              "[": LLBracket,
              "]": LRBracket,
              "(": LLRBracket,
              ")": LRRBracket,
              "!": LNotCond}


tokens_oper = {"": LSet,
               "~": LLazySet,
               "+": LAppend,
               "<": LPrepend,
               "?": LRegExpSet,
               "?+": LRegExpAppend,
               "?<": LRegExpPrepend,
               }


tokens_oper_re = [r"\=", r"\+\=", r"\<\=", r"\~\=", r"\?\=", r"\?\+\=", r"\?\<\="]


_ops_exp = re.compile(r"|".join(tokens_oper_re))


class Lexer(object):

    def __init__(self, reader):
        self.reader = reader
        self.filename = reader.filename
        self.line = None
        self.linenum = 0
        self.ignore_white = False
        self.rest_as_string = False
        self.match_func_index = 0
        self.generator = self.get_lexer()
        self.prev_indent = 0
        self.fast = False

    def set_prev_indent(self, prev_indent):
        self.prev_indent = prev_indent

    def set_fast(self):
        self.fast = True

    def set_strict(self):
        self.fast = False

    def match(self, line, pos):
        l0 = line[0]
        chars = ""
        m = None
        cind = 0
        if l0 == "v":
            if line.startswith("variants:"):
                yield LVariants()
                yield LColon()
                pos = 9
            elif line.startswith("variants "):
                yield LVariants()
                pos = 8
        elif l0 == "-":
            yield LVariant()
            pos = 1
        elif l0 == "o":
            if line.startswith("only "):
                yield LOnly()
                pos = 4
                while line[pos].isspace():
                    pos += 1
        elif l0 == "n":
            if line.startswith("no "):
                yield LNo()
                pos = 2
                while line[pos].isspace():
                    pos += 1
        elif l0 == "i":
            if line.startswith("include "):
                yield LInclude()
                pos = 7
        elif l0 == "d":
            if line.startswith("del "):
                yield LDel()
                pos = 3
                while line[pos].isspace():
                    pos += 1
        elif l0 == "s":
            if line.startswith("suffix "):
                yield LSuffix()
                pos = 6
                while line[pos].isspace():
                    pos += 1
        elif l0 == "j":
            if line.startswith("join "):
                yield LJoin()
                pos = 4
                while line[pos].isspace():
                    pos += 1

        if self.fast and pos == 0:  # due to refexp
            cind = line[pos:].find(":")
            m = _ops_exp.search(line[pos:])

        oper = ""
        token = None

        if self.rest_as_string:
            self.rest_as_string = False
            yield LString(line[pos:].lstrip())
        elif self.fast and m and (cind < 0 or cind > m.end()):
            chars = ""
            yield LIdentifier(line[:m.start()].rstrip())
            yield tokens_oper[m.group()[:-1]]()
            yield LString(line[m.end():].lstrip())
        else:
            li = enum(line[pos:], pos)
            for pos, char in li:
                if char.isalnum() or char in spec_iden:    # alfanum+_-
                    chars += char
                elif char in spec_oper:     # <+?=~
                    if chars:
                        yield LIdentifier(chars)
                        oper = ""
                    chars = ""
                    oper += char
                else:
                    if chars:
                        yield LIdentifier(chars)
                        chars = ""
                    if char.isspace():   # Whitespace
                        for pos, char in li:
                            if not char.isspace():
                                if not self.ignore_white:
                                    yield LWhite()
                                break
                    if char.isalnum() or char in spec_iden:
                        chars += char
                    elif char == "=":
                        if oper in tokens_oper:
                            yield tokens_oper[oper]()
                        else:
                            raise LexerError("Unexpected character %s on"
                                             " pos %s" % (char, pos),
                                             self.line, self.filename,
                                             self.linenum)
                        oper = ""
                    elif char in tokens_map:
                        token = tokens_map[char]()
                    elif char == "\"":
                        chars = ""
                        pos, char = next(li)
                        while char != "\"":
                            chars += char
                            pos, char = next(li)
                        yield LString(chars)
                    elif char == "#":
                        break
                    elif char in spec_oper:
                        oper += char
                    else:
                        raise LexerError("Unexpected character %s on"
                                         " pos %s. Special chars are allowed"
                                         " only in variable assignation"
                                         " statement" % (char, pos), line,
                                         self.filename, self.linenum)
                    if token is not None:
                        yield token
                        token = None
                    if self.rest_as_string:
                        self.rest_as_string = False
                        yield LString(line[pos + 1:].lstrip())
                        break
        if chars:
            yield LIdentifier(chars)
            chars = ""
        yield LEndL()

    def get_lexer(self):
        cr = self.reader
        indent = 0
        while True:
            (self.line, indent,
             self.linenum) = cr.get_next_line(self.prev_indent)

            if not self.line:
                yield LEndBlock(indent)
                continue

            yield LIndent(indent)
            for token in self.match(self.line, 0):
                yield token

    def get_until_gen(self, end_tokens=None):
        if end_tokens is None:
            end_tokens = [LEndL]
        token = next(self.generator)
        while type(token) not in end_tokens:
            yield token
            token = next(self.generator)
        yield token

    def get_until(self, end_tokens=None):
        if end_tokens is None:
            end_tokens = [LEndL]
        return [x for x in self.get_until_gen(end_tokens)]

    def flush_until(self, end_tokens=None):
        if end_tokens is None:
            end_tokens = [LEndL]
        for _ in self.get_until_gen(end_tokens):
            pass

    def get_until_check(self, lType, end_tokens=None):
        """
        Read tokens from iterator until get end_tokens or type of token not
        match ltype

        :param lType: List of allowed tokens
        :param end_tokens: List of tokens for end reading
        :return: List of readed tokens.
        """
        if end_tokens is None:
            end_tokens = [LEndL]
        tokens = []
        lType = lType + end_tokens
        for token in self.get_until_gen(end_tokens):
            if type(token) in lType:
                tokens.append(token)
            else:
                raise ParserError("Expected %s got %s" % (lType, type(token)),
                                  self.line, self.filename, self.linenum)
        return tokens

    def get_until_no_white(self, end_tokens=None):
        """
        Read tokens from iterator until get one of end_tokens and strip LWhite

        :param end_tokens:  List of tokens for end reading
        :return: List of readed tokens.
        """
        if end_tokens is None:
            end_tokens = [LEndL]
        return [x for x in self.get_until_gen(end_tokens) if type(x) != LWhite]

    def rest_line_gen(self):
        token = next(self.generator)
        while type(token) != LEndL:
            yield token
            token = next(self.generator)

    def rest_line(self):
        return [x for x in self.rest_line_gen()]

    def rest_line_no_white(self):
        return [x for x in self.rest_line_gen() if type(x) != LWhite]

    def rest_line_as_LString(self):
        self.rest_as_string = True
        lstr = next(self.generator)
        next(self.generator)
        return lstr

    def get_next_check(self, lType):
        token = next(self.generator)
        if type(token) in lType:
            return type(token), token
        else:
            raise ParserError("Expected %s got ['%s']=[%s]" %
                              ([x.identifier for x in lType],
                               token.identifier, token),
                              self.line, self.filename, self.linenum)

    def get_next_check_nw(self, lType):
        token = next(self.generator)
        while type(token) == LWhite:
            token = next(self.generator)
        if type(token) in lType:
            return type(token), token
        else:
            raise ParserError("Expected %s got ['%s']" %
                              ([x.identifier for x in lType],
                               token.identifier),
                              self.line, self.filename, self.linenum)

    def check_token(self, token, lType):
        if type(token) in lType:
            return type(token), token
        else:
            raise ParserError("Expected %s got ['%s']" %
                              ([x.identifier for x in lType],
                               token.identifier),
                              self.line, self.filename, self.linenum)


def next_nw(gener):
    token = next(gener)
    while type(token) == LWhite:
        token = next(gener)
    return token


def cmd_tokens(tokens1, tokens2):
    for x, y in list(zip(tokens1, tokens2)):
        if x != y:
            return False
    else:
        return True


def apply_predict(lexer, node, pre_dict):
    predict = LApplyPreDict().set_operands(None, pre_dict)
    node.content += [(lexer.filename, lexer.linenum, predict)]
    return {}


def parse_filter(lexer, tokens):
    """
    :return: Parsed filter
    """
    or_filters = []
    tokens = iter(tokens + [LEndL()])
    typet, token = lexer.check_token(next(tokens), [LIdentifier, LLRBracket,
                                                    LEndL, LWhite])
    and_filter = []
    con_filter = []
    dots = 1
    while typet not in [LEndL]:
        if typet in [LIdentifier, LLRBracket]:        # join    identifier
            if typet == LLRBracket:    # (xxx=ttt)
                _, ident = lexer.check_token(next_nw(tokens),
                                             [LIdentifier])  # (iden
                typet, _ = lexer.check_token(next_nw(tokens),
                                             [LSet, LRRBracket])  # =
                if typet == LRRBracket:  # (xxx)
                    token = Label(str(ident))
                elif typet == LSet:    # (xxx = yyyy)
                    _, value = lexer.check_token(next_nw(tokens),
                                                 [LIdentifier, LString])
                    lexer.check_token(next_nw(tokens), [LRRBracket])
                    token = Label(str(ident), str(value))
            else:
                token = Label(token)
            if dots == 1:
                con_filter.append(token)
            elif dots == 2:
                and_filter.append(con_filter)
                con_filter = [token]
            elif dots == 0 or dots > 2:
                raise ParserError("Syntax Error expected \".\" between"
                                  " Identifier.", lexer.line, lexer.filename,
                                  lexer.linenum)

            dots = 0
        elif typet == LDot:         # xxx.xxxx or xxx..xxxx
            dots += 1
        elif typet in [LComa, LWhite]:
            if dots > 0:
                raise ParserError("Syntax Error expected identifier between"
                                  " \".\" and \",\".", lexer.line,
                                  lexer.filename, lexer.linenum)
            if and_filter:
                if con_filter:
                    and_filter.append(con_filter)
                    con_filter = []
                or_filters.append(and_filter)
                and_filter = []
            elif con_filter:
                or_filters.append([con_filter])
                con_filter = []
            elif typet == LIdentifier:
                or_filters.append([[Label(token)]])
            else:
                raise ParserError("Syntax Error expected \",\" between"
                                  " Identifier.", lexer.line, lexer.filename,
                                  lexer.linenum)
            dots = 1
            token = next(tokens)
            while type(token) == LWhite:
                token = next(tokens)
            typet, token = lexer.check_token(token, [LIdentifier,
                                                     LComa, LDot,
                                                     LLRBracket, LEndL])
            continue
        typet, token = lexer.check_token(next(tokens), [LIdentifier, LComa,
                                                        LDot, LLRBracket,
                                                        LEndL, LWhite])
    if and_filter:
        if con_filter:
            and_filter.append(con_filter)
            con_filter = []
        or_filters.append(and_filter)
        and_filter = []
    if con_filter:
        or_filters.append([con_filter])
        con_filter = []
    return or_filters


class Parser(object):
    # pylint: disable=W0102

    def __init__(self, filename=None, defaults=False, expand_defaults=[],
                 debug=False):
        self.node = Node()
        self.debug = debug
        self.defaults = defaults
        self.expand_defaults = [LIdentifier(x) for x in expand_defaults]

        self.filename = filename
        if self.filename:
            self.parse_file(self.filename)

        self.only_filters = []
        self.no_filters = []
        self.assignments = []

        # get_dicts() - is recursive generator, it can invoke itself,
        # as well as it can be called outside to get dic list
        # It is necessary somehow mark top-level generator,
        # to be able process all variables, do suffix stuff, drops dups, etc....
        # It can be safely done only on top top level get_dicts()
        # Parent generator will reset this flag
        self.parent_generator = True

    def _debug(self, s, *args):
        if self.debug:
            logging.debug(s, *args)

    def _warn(self, s, *args):
        logging.warn(s, *args)

    def parse_file(self, filename):
        """
        Parse a file.

        :param filename: Path of the configuration file.
        """
        self.node.filename = filename
        self.node = self._parse(Lexer(FileReader(filename)), self.node)
        self.filename = filename

    def parse_string(self, s):
        """
        Parse a string.

        :param s: String to parse.
        """
        self.node.filename = StrReader("").filename
        self.node = self._parse(Lexer(StrReader(s)), self.node)

    def only_filter(self, variant):
        """
        Apply a only filter programatically and keep track of it.

        Equivalent to parse a "only variant" line.

        :param variant: String with the variant name.
        """
        string = "only %s" % variant
        self.only_filters.append(string)
        self.parse_string(string)

    def no_filter(self, variant):
        """
        Apply a no filter programatically and keep track of it.

        Equivalent to parse a "no variant" line.

        :param variant: String with the variant name.
        """
        string = "no %s" % variant
        self.no_filters.append(string)
        self.parse_string(string)

    def assign(self, key, value):
        """
        Apply an assigment programatically and keep track of it.

        Equivalent to parse a "key = value" line.

        :param variant: String with the variant name.
        """
        string = "%s = %s" % (key, value)
        self.assignments.append(string)
        self.parse_string(string)

    def _parse(self, lexer, node=None, prev_indent=-1):
        if not node:
            node = self.node
        block_allowed = [LVariants, LIdentifier, LOnly,
                         LNo, LInclude, LDel, LNotCond, LSuffix, LJoin]

        variants_allowed = [LVariant]

        identifier_allowed = [LSet, LAppend, LPrepend, LLazySet,
                              LRegExpSet, LRegExpAppend,
                              LRegExpPrepend, LColon,
                              LEndL]

        varianst_allowed_in = [LLBracket, LColon, LIdentifier, LEndL]
        indent_allowed = [LIndent, LEndBlock]

        allowed = block_allowed
        var_indent = 0
        var_name = ""
        # meta contains variants meta-data
        meta = {}
        # pre_dict contains block of operation without collision with
        # others block or operation. Increase speed almost twice.
        pre_dict = {}
        lexer.set_fast()

        # Suffix should be applied as the last operator in the dictionary
        # Reasons:
        #     1. Escape multiply suffix operators
        #     2. Affect all elements in current block
        suffix = None

        try:
            while True:
                lexer.set_prev_indent(prev_indent)
                typet, token = lexer.get_next_check(indent_allowed)
                if typet == LEndBlock:
                    if pre_dict:
                        # flush pre_dict to node content.
                        pre_dict = apply_predict(lexer, node, pre_dict)
                    if suffix:
                        # Node has suffix, apply it to all elements
                        node.content.append(suffix)
                    return node

                indent = token.length
                typet, token = lexer.get_next_check(allowed)

                if typet == LIdentifier:
                    # Parse:
                    #    identifier .....
                    identifier = lexer.get_until_no_white(identifier_allowed)
                    if isinstance(identifier[-1], LOperators):  # operand = <=
                        # Parse:
                        #    identifier = xxx
                        #    identifier <= xxx
                        #    identifier ?= xxx
                        #    etc..
                        op = identifier[-1]
                        if (len(identifier) == 1):
                            identifier = token
                        else:
                            identifier = [token] + identifier[:-1]
                            identifier = "".join([str(x) for x in identifier])
                        _, value = lexer.get_next_check([LString])
                        if value and (value[0] == value[-1] == '"' or
                                      value[0] == value[-1] == "'"):
                            value = value[1:-1]

                        op.set_operands(identifier, value)
                        d_nin_val = "$" not in value
                        if type(op) == LSet and d_nin_val:  # Optimization
                            op.apply_to_dict(pre_dict)
                        else:
                            if pre_dict:
                                # flush pre_dict to node content.
                                # If block already contain xxx = yyyy
                                # then operation xxx +=, <=, .... are safe.
                                if op.name in pre_dict and d_nin_val:
                                    op.apply_to_dict(pre_dict)
                                    lexer.get_next_check([LEndL])
                                    continue
                                else:
                                    pre_dict = apply_predict(lexer, node,
                                                             pre_dict)

                            node.content += [(lexer.filename,
                                              lexer.linenum,
                                              op)]
                        lexer.get_next_check([LEndL])

                    elif type(identifier[-1]) == LColon:  # condition:
                        # Parse:
                        #    xxx.yyy.(aaa=bbb):
                        identifier = [token] + identifier[:-1]
                        cfilter = parse_filter(lexer, identifier + [LEndL()])
                        next_line = lexer.rest_line_as_LString()
                        if next_line != "":
                            lexer.reader.set_next_line(next_line, indent + 1,
                                                       lexer.linenum)
                        cond = Condition(cfilter, lexer.line)
                        self._parse(lexer, cond, prev_indent=indent)

                        pre_dict = apply_predict(lexer, node, pre_dict)
                        node.content += [(lexer.filename, lexer.linenum, cond)]
                    else:
                        raise ParserError("Syntax ERROR expected \":\" or"
                                          " operand", lexer.line,
                                          lexer.filename, lexer.linenum)

                elif typet == LVariant:
                    # Parse
                    #  - var1: depend1, depend2
                    #      block1
                    #  - var2:
                    #      block2
                    if pre_dict:
                        pre_dict = apply_predict(lexer, node, pre_dict)
                    already_default = False
                    is_default = False
                    meta_with_default = False
                    if "default" in meta:
                        meta_with_default = True
                    meta_in_expand_defautls = False
                    if var_name not in self.expand_defaults:
                        meta_in_expand_defautls = True
                    node4 = Node()
                    while True:
                        lexer.set_prev_indent(var_indent)
                        # Get token from lexer and check syntax.
                        typet, token = lexer.get_next_check_nw([LIdentifier,
                                                                LDefault,
                                                                LIndent,
                                                                LEndBlock])
                        if typet == LEndBlock:
                            break

                        if typet == LIndent:
                            lexer.get_next_check_nw([LVariant])
                            typet, token = lexer.get_next_check_nw(
                                [LIdentifier,
                                 LDefault])

                        if typet == LDefault:  # @
                            is_default = True
                            name = lexer.get_until_check([LIdentifier, LDot],
                                                         [LColon])
                        else:  # identificator
                            is_default = False
                            name = [token] + lexer.get_until_check(
                                [LIdentifier, LDot],
                                [LColon])

                        if len(name) == 2:
                            name = [name[0]]
                            raw_name = name
                        else:
                            raw_name = [x for x in name[:-1]]
                            name = [x for x in name[:-1]
                                    if type(x) == LIdentifier]

                        token = next(lexer.generator)
                        while type(token) == LWhite:
                            token = next(lexer.generator)
                        tokens = None
                        if type(token) != LEndL:
                            tokens = [token] + lexer.get_until([LEndL])
                            deps = parse_filter(lexer, tokens)
                        else:
                            deps = []

                        # Prepare data for dict generator.
                        node2 = Node()
                        node2.children = [node]
                        node2.labels = node.labels

                        if var_name:
                            op = LSet().set_operands(var_name,
                                                     ".".join([str(n) for n in name]))
                            node2.content += [(lexer.filename,
                                               lexer.linenum,
                                               op)]

                        node3 = self._parse(lexer, node2, prev_indent=indent)

                        if var_name:
                            node3.var_name = var_name
                            node3.name = [Label(var_name, str(n))
                                          for n in name]
                        else:
                            node3.name = [Label(str(n)) for n in name]

                        # Update mapping name to file

                        node3.dep = deps

                        if meta_with_default:
                            for wd in meta["default"]:
                                if cmd_tokens(wd, raw_name):
                                    is_default = True
                                    meta["default"].remove(wd)

                        if (is_default and not already_default and
                                meta_in_expand_defautls):
                            node3.default = True
                            already_default = True

                        node3.append_to_shortname = not is_default

                        op = LUpdateFileMap()
                        op.set_operands(lexer.filename,
                                        ".".join(str(x)
                                                 for x in node3.name))
                        node3.content += [(lexer.filename,
                                           lexer.linenum,
                                           op)]

                        op = LUpdateFileMap()
                        op.set_operands(lexer.filename,
                                        ".".join(str(x.name)
                                                 for x in node3.name),
                                        "_short_name_map_file")
                        node3.content += [(lexer.filename,
                                           lexer.linenum,
                                           op)]

                        if node3.default and self.defaults:
                            # Move default variant in front of rest
                            # of all variants.
                            # Speed optimization.
                            node4.children.insert(0, node3)
                        else:
                            node4.children += [node3]
                        node4.labels.update(node3.labels)
                        node4.labels.update(node3.name)

                    if "default" in meta and meta["default"]:
                        raise ParserError("Missing default variant %s" %
                                          (meta["default"]), lexer.line,
                                          lexer.filename, lexer.linenum)
                    allowed = block_allowed
                    node = node4

                elif typet == LVariants:  # _name_ [meta1=xxx] [yyy] [xxx]
                    # Parse
                    #    variants _name_ [meta1] [meta2]:
                    if type(node) in [Condition, NegativeCondition]:
                        raise ParserError("'variants' is not allowed inside a "
                                          "conditional block", lexer.line,
                                          lexer.reader.filename, lexer.linenum)

                    lexer.set_strict()
                    tokens = lexer.get_until_no_white([LLBracket, LColon,
                                                       LIdentifier, LEndL])
                    vtypet = type(tokens[-1])
                    var_name = ""
                    meta.clear()
                    # [meta1=xxx] [yyy] [xxx]
                    while vtypet not in [LColon, LEndL]:
                        if vtypet == LIdentifier:
                            if var_name != "":
                                raise ParserError("Syntax ERROR expected"
                                                  " \"[\" or \":\"",
                                                  lexer.line, lexer.filename,
                                                  lexer.linenum)
                            var_name = tokens[0]
                        elif vtypet == LLBracket:  # [
                            _, ident = lexer.get_next_check_nw([LIdentifier])
                            typet, _ = lexer.get_next_check_nw([LSet,
                                                                LRBracket])
                            if typet == LRBracket:  # [xxx]
                                if ident not in meta:
                                    meta[ident] = []
                                meta[ident].append(True)
                            elif typet == LSet:  # [xxx = yyyy]
                                tokens = lexer.get_until_no_white([LRBracket,
                                                                   LEndL])
                                if type(tokens[-1]) == LRBracket:
                                    if ident not in meta:
                                        meta[ident] = []
                                    meta[ident].append(tokens[:-1])
                                else:
                                    raise ParserError("Syntax ERROR"
                                                      " expected \"]\"",
                                                      lexer.line,
                                                      lexer.filename,
                                                      lexer.linenum)

                        tokens = lexer.get_next_check_nw(varianst_allowed_in)
                        vtypet = type(tokens[-1])

                    if "default" in meta:
                        for wd in meta["default"]:
                            if type(wd) != list:
                                raise ParserError("Syntax ERROR expected "
                                                  "[default=xxx]",
                                                  lexer.line,
                                                  lexer.filename,
                                                  lexer.linenum)

                    if vtypet == LEndL:
                        raise ParserError("Syntax ERROR expected \":\"",
                                          lexer.line, lexer.filename,
                                          lexer.linenum)
                    lexer.get_next_check_nw([LEndL])
                    allowed = variants_allowed
                    var_indent = indent

                elif typet in [LNo, LOnly]:
                    # Parse:
                    #    only/no (filter=text)..aaa.bbb, xxxx
                    lfilter = parse_filter(lexer, lexer.rest_line())

                    pre_dict = apply_predict(lexer, node, pre_dict)
                    if typet == LOnly:
                        node.content += [(lexer.filename, lexer.linenum,
                                          OnlyFilter(lfilter, lexer.line))]
                    else:  # LNo
                        node.content += [(lexer.filename, lexer.linenum,
                                          NoFilter(lfilter, lexer.line))]

                elif typet == LJoin:
                    # Parse:
                    #    join (filter=text)..aaa.bbb, xxxx
                    # syntax is the same as for No/Only filters
                    lfilter = parse_filter(lexer, lexer.rest_line())

                    pre_dict = apply_predict(lexer, node, pre_dict)

                    node.content += [(lexer.filename, lexer.linenum, JoinFilter(lfilter, lexer.line))]

                elif typet == LSuffix:
                    # Parse:
                    #    suffix SUFFIX
                    if pre_dict:
                        pre_dict = apply_predict(lexer, node, pre_dict)
                    token_type, token_val = lexer.get_next_check([LIdentifier])
                    lexer.get_next_check([LEndL])
                    suffix_operator = Suffix().set_operands(None, token_val)
                    # Suffix will be applied as all other elements in current node are processed:
                    suffix = (lexer.filename, lexer.linenum, suffix_operator)

                elif typet == LInclude:
                    # Parse:
                    #    include relative file patch to working directory.
                    path = lexer.rest_line_as_LString()
                    filename = os.path.expanduser(path)
                    if (isinstance(lexer.reader, FileReader) and
                            not os.path.isabs(filename)):
                        filename = os.path.join(
                            os.path.dirname(lexer.filename),
                            filename)
                    if not os.path.isfile(filename):
                        raise MissingIncludeError(lexer.line, lexer.filename,
                                                  lexer.linenum)
                    pre_dict = apply_predict(lexer, node, pre_dict)
                    lch = Lexer(FileReader(filename))
                    node = self._parse(lch, node, -1)
                    lexer.set_prev_indent(prev_indent)

                elif typet == LDel:
                    # Parse:
                    #    del operand
                    _, to_del = lexer.get_next_check_nw([LIdentifier])
                    lexer.get_next_check_nw([LEndL])
                    token.set_operands(to_del, None)

                    pre_dict = apply_predict(lexer, node, pre_dict)
                    node.content += [(lexer.filename, lexer.linenum,
                                      token)]

                elif typet == LNotCond:
                    # Parse:
                    #    !xxx.yyy.(aaa=bbb): vvv
                    lfilter = parse_filter(lexer,
                                           lexer.get_until_no_white(
                                               [LColon, LEndL])[:-1])
                    next_line = lexer.rest_line_as_LString()
                    if next_line != "":
                        lexer.reader.set_next_line(next_line, indent + 1,
                                                   lexer.linenum)
                    cond = NegativeCondition(lfilter, lexer.line)
                    self._parse(lexer, cond, prev_indent=indent)
                    lexer.set_prev_indent(prev_indent)

                    pre_dict = apply_predict(lexer, node, pre_dict)
                    node.content += [(lexer.filename, lexer.linenum, cond)]
                else:
                    raise ParserError("Syntax ERROR expected", lexer.line,
                                      lexer.filename, lexer.linenum)
        except Exception:
            self._debug("%s  %s:  %s" % (lexer.filename, lexer.linenum,
                                         lexer.line))
            raise

    def drop_suffixes(self, d):
        """
        Merge suffixes for same var, or drop off unnecessary suffixes

        This step can be done safely only by the top level generator, before
        outputting the dictionary to outside world.
        """
        # dictionary `d' is going to change, keep its original copy
        d_orig = d.copy()
        for key in d_orig:
            if key in _reserved_keys:
                continue

            if not isinstance(key, tuple):
                continue

            try:
                # This file was invoked through cmdline
                options.skipdups
                skipdups = options.skipdups
            except NameError:
                # This file was invoked as Python module
                skipdups = True

            if skipdups:
                # Drop vars with suffixes matches general var val
                # Example: if a_x == 1, and a == 1. Drop: a_x, leave a
                gen_var_name = key[0]
                if gen_var_name in d_orig and d_orig[gen_var_name] == d_orig[key]:
                    # Drop gen_var_name, use general key with same value
                    d.pop(key)
                    continue

                can_drop_all_suffixes_for_this_key = True
                for k in d_orig:
                    gen_name = k[0] if isinstance(k, tuple) else k
                    if gen_var_name == gen_name:
                        if d_orig[key] != d_orig[k]:
                            can_drop_all_suffixes_for_this_key = False
                            break

            if skipdups and can_drop_all_suffixes_for_this_key:
                new_key = key[0]
            else:
                # merge suffixes, preserve reverse order of suffixes
                new_key = key[:1] + key[1:][::-1]
                new_key = ''.join((map(str, new_key)))
            d[new_key] = d.pop(key)

    def get_dicts(self, node=None, ctx=[], content=[], shortname=[], dep=[]):
        """
        Process 'join' entry, unpack join filter for node.

        :param ctx: node labels/names
        :param content: previous content in plain
        :returns: dictionary

        1) join filter_1 filter_2 ....
            multiplies all dictionaries as:
                all_variants_match_filter_1 * all_variants_match_filter_2 * ....
        2) join only_one_filter
                == only only_one_filter
        3) join filter_1 filter_1
            also works and transforms to:
                all_variants_match_filter_1 * all_variants_match_filter_1
            Example:
                join a
                join a
            Transforms into:
                join a a
        """
        node = node or self.node

        # Keep track to know who is a parent generator
        parent = False
        if self.parent_generator:
            # I am parent of the all
            parent = True
            # No one else is
            self.parent_generator = False

        # Node is a current block. It has content, its contents: node.content
        # Content without joins
        new_content = []

        # All joins in current node
        joins = []

        for t in node.content:
            filename, linenum, obj = t

            if not isinstance(obj, JoinFilter):
                new_content.append(t)
                continue

            # Accumulate all joins at one node
            joins += [t]

        if not joins:
            # Return generator
            for d in self.get_dicts_plain(node, ctx, content, shortname, dep):
                if parent:
                    self.drop_suffixes(d)
                yield d
        else:
            # Rewrite all separate joins in one node as many `only'
            onlys = []
            for j in joins:
                filename, linenum, obj = j
                for word in obj.filter:
                    f = OnlyFilter([word], str(word))
                    onlys += [(filename, linenum, f)]

            old_content = node.content[:]
            node.content = new_content
            for d in self.multiply_join(onlys, node, ctx, content, shortname, dep):
                if parent:
                    self.drop_suffixes(d)
                yield d
            node.content = old_content[:]

    def mk_name(self, n1, n2):
        """Make name for test. Case: two dics were merged"""
        common_prefix = n1[:[x[0] == x[1] for x in list(zip(n1, n2))].index(0)]
        cp = ".".join(common_prefix.split('.')[:-1])
        p1 = re.sub(r"^"+cp, "", n1)
        p2 = re.sub(r"^"+cp, "", n2)
        if cp:
            name = cp + p1 + p2
        else:
            name = p1 + "." + p2
        return name

    def multiply_join(self, onlys, node=None, ctx=[], content=[], shortname=[], dep=[]):
        """
        Multiply all joins. Return dictionaries one by one
        Each `join' is the same as `only' filter
        This functions is supposed to be a generator, recursive generator
        """
        # Current join/only
        only = onlys[:1]
        remains = onlys[1:]

        content_orig = node.content[:]
        node.content += only

        if not remains:
            for d in self.get_dicts_plain(node, ctx, content, shortname, dep):
                yield d
        else:
            for d1 in self.get_dicts_plain(node, ctx, content, shortname, dep):
                # Current frame multiply by all variants from bottom
                node.content = content_orig
                for d2 in self.multiply_join(remains, node, ctx, content, shortname, dep):

                    d = d1.copy()
                    d.update(d2)
                    d["name"] = self.mk_name(d1["name"], d2["name"])
                    d["shortname"] = self.mk_name(d1["shortname"], d2["shortname"])
                    yield d

    def get_dicts_plain(self, node=None, ctx=[], content=[], shortname=[], dep=[]):
        """
        Generate dictionaries from the code parsed so far.  This should
        be called after parsing something.

        :return: A dict generator.
        """
        def process_content(content, failed_filters):
            # 1. Check that the filters in content are OK with the current
            #    context (ctx).
            # 2. Move the parts of content that are still relevant into
            #    new_content and unpack conditional blocks if appropriate.
            #    For example, if an 'only' statement fully matches ctx, it
            #    becomes irrelevant and is not appended to new_content.
            #    If a conditional block fully matches, its contents are
            #    unpacked into new_content.
            # 3. Move failed filters into failed_filters, so that next time we
            #    reach this node or one of its ancestors, we'll check those
            #    filters first.
            blocked_filters = []
            for t in content:
                filename, linenum, obj = t
                if isinstance(obj, LOperators):
                    new_content.append(t)
                    continue
                # obj is an OnlyFilter/NoFilter/Condition/NegativeCondition
                if obj.requires_action(ctx, ctx_set, labels):
                    # This filter requires action now
                    if type(obj) is OnlyFilter or type(obj) is NoFilter:
                        if obj not in blocked_filters:
                            self._debug("    filter did not pass: %r (%s:%s)",
                                        obj.line, filename, linenum)
                            failed_filters.append(t)
                            return False
                        else:
                            continue
                    else:
                        self._debug("    conditional block matches:"
                                    " %r (%s:%s)", obj.line, filename, linenum)
                        # Check and unpack the content inside this Condition
                        # object (note: the failed filters should go into
                        # new_internal_filters because we don't expect them to
                        # come from outside this node, even if the Condition
                        # itself was external)
                        if not process_content(obj.content,
                                               new_internal_filters):
                            failed_filters.append(t)
                            return False
                        continue
                elif obj.is_irrelevant(ctx, ctx_set, labels):
                    # This filter is no longer relevant and can be removed
                    continue
                else:
                    # Keep the filter and check it again later
                    new_content.append(t)
            return True

        def might_pass(failed_ctx,
                       failed_ctx_set,
                       failed_external_filters,
                       failed_internal_filters):
            all_content = content + node.content
            for t in failed_external_filters + failed_internal_filters:
                if t not in all_content:
                    return True
            for t in failed_external_filters:
                _, _, external_filter = t
                if not external_filter.might_pass(failed_ctx,
                                                  failed_ctx_set,
                                                  ctx, ctx_set,
                                                  labels):
                    return False
            for t in failed_internal_filters:
                if t not in node.content:
                    return True

            for t in failed_internal_filters:
                _, _, internal_filter = t
                if not internal_filter.might_pass(failed_ctx,
                                                  failed_ctx_set,
                                                  ctx, ctx_set,
                                                  labels):
                    return False
            return True

        def add_failed_case():
            node.failed_cases.appendleft((ctx, ctx_set,
                                          new_external_filters,
                                          new_internal_filters))
            if len(node.failed_cases) > num_failed_cases:
                node.failed_cases.pop()

        node = node or self.node
        # if self.debug:    #Print dict on which is working now.
        #    node.dump(0)
        # Update dep
        for d in node.dep:
            for dd in d:
                dep = dep + [".".join([str(label) for label in ctx + dd])]
        # Update ctx
        ctx = ctx + node.name
        ctx_set = set(ctx)
        labels = node.labels
        # Get the current name
        name = ".".join([str(label) for label in ctx])

        if node.name:
            self._debug("checking out %r", name)

        # Check previously failed filters
        for i, failed_case in enumerate(node.failed_cases):
            if not might_pass(*failed_case):
                self._debug("\n*    this subtree has failed before %s\n"
                            "         content: %s\n"
                            "         failcase:%s\n",
                            name, content + node.content, failed_case)
                del node.failed_cases[i]
                node.failed_cases.appendleft(failed_case)
                return

        # Check content and unpack it into new_content
        new_content = []
        new_external_filters = []
        new_internal_filters = []
        if (not process_content(node.content, new_internal_filters) or
                not process_content(content, new_external_filters)):
            add_failed_case()
            self._debug("Failed_cases %s", node.failed_cases)
            return

        # Update shortname
        if node.append_to_shortname:
            shortname = shortname + node.name

        # Recurse into children
        count = 0
        if self.defaults and node.var_name not in self.expand_defaults:
            for n in node.children:
                for d in self.get_dicts(n, ctx, new_content, shortname, dep):
                    count += 1
                    yield d
                if n.default and count:
                    break
        else:
            for n in node.children:
                for d in self.get_dicts(n, ctx, new_content, shortname, dep):
                    count += 1
                    yield d
        # Reached leaf?
        if not node.children:
            self._debug("    reached leaf, returning it")
            d = {"name": name, "dep": dep,
                 "shortname": ".".join([str(sn.name) for sn in shortname])}
            for _, _, op in new_content:
                op.apply_to_dict(d)
            postfix_parse(d)
            yield d


def print_dicts_default(options, dicts):
    """Print dictionaries in the default mode"""
    for count, dic in enumerate(dicts):
        if options.fullname:
            print("dict %4d:  %s" % (count + 1, dic["name"]))
        else:
            print("dict %4d:  %s" % (count + 1, dic["shortname"]))
        if options.contents:
            keys = list(dic.keys())
            keys.sort()
            for key in keys:
                print("    %s = %s" % (key, dic[key]))


# pylint: disable=W0613
def print_dicts_repr(options, dicts):
    import pprint
    print("[")
    for dic in dicts:
        print("%s," % (pprint.pformat(dic)))
    print("]")


def print_dicts(options, dicts):
    if options.repr_mode:
        print_dicts_repr(options, dicts)
    else:
        print_dicts_default(options, dicts)


def convert_data_size(size, default_sufix='B'):
    """
    Convert data size from human readable units to an int of arbitrary size.

    :param size: Human readable data size representation (string).
    :param default_sufix: Default sufix used to represent data.
    :return: Int with data size in the appropriate order of magnitude.
    """
    orders = {'B': 1,
              'K': 1024,
              'M': 1024 * 1024,
              'G': 1024 * 1024 * 1024,
              'T': 1024 * 1024 * 1024 * 1024,
              }

    order = re.findall("([BbKkMmGgTt])", size[-1])
    if not order:
        size += default_sufix
        order = [default_sufix]

    return int(float(size[0:-1]) * orders[order[0].upper()])


def compare_string(str1, str2):
    """
    Compare two int string and return -1, 0, 1.
    It can compare two memory value even in sufix

    :param str1: The first string
    :param str2: The second string

    :Return: Rteurn -1, when str1<  str2
                     0, when str1 = str2
                     1, when str1>  str2
    """
    order1 = re.findall("([BbKkMmGgTt])", str1)
    order2 = re.findall("([BbKkMmGgTt])", str2)
    if order1 or order2:
        value1 = convert_data_size(str1, "M")
        value2 = convert_data_size(str2, "M")
    else:
        value1 = int(str1)
        value2 = int(str2)
    if value1 < value2:
        return -1
    elif value1 == value2:
        return 0
    else:
        return 1


def postfix_parse(dic):
    tmp_dict = {}
    for key in dic:
        # Bypass the case that use tuple as key value
        if isinstance(key, tuple):
            continue
        if key.endswith("_max"):
            tmp_key = key.split("_max")[0]
            if (tmp_key not in dic or
                    compare_string(dic[tmp_key], dic[key]) > 0):
                tmp_dict[tmp_key] = dic[key]
        elif key.endswith("_min"):
            tmp_key = key.split("_min")[0]
            if (tmp_key not in dic or
                    compare_string(dic[tmp_key], dic[key]) < 0):
                tmp_dict[tmp_key] = dic[key]
        elif key.endswith("_fixed"):
            tmp_key = key.split("_fixed")[0]
            tmp_dict[tmp_key] = dic[key]
    for key in tmp_dict:
        dic[key] = tmp_dict[key]


if __name__ == "__main__":
    parser = optparse.OptionParser('usage: %prog [options] filename '
                                   '[extra code] ...\n\nExample:\n\n    '
                                   '%prog tests.cfg "only my_set" "no qcow2"')
    parser.add_option("-v", "--verbose", dest="debug", action="store_true",
                      help="include debug messages in console output")
    parser.add_option("-f", "--fullname", dest="fullname", action="store_true",
                      help="show full dict names instead of short names")
    parser.add_option("-c", "--contents", dest="contents", action="store_true",
                      help="show dict contents")
    parser.add_option("-r", "--repr", dest="repr_mode", action="store_true",
                      help="output parsing results Python format")
    parser.add_option("-d", "--defaults", dest="defaults", action="store_true",
                      help="use only default variant of variants if there"
                           " is some")
    parser.add_option("-e", "--expand", dest="expand", type="string",
                      help="list of vartiant which should be expanded when"
                           " defaults is enabled.  \"name, name, name\"")
    parser.add_option("-s", "--skip-dups", dest="skipdups", default=True, action="store_false",
                      help="Don't drop variables with different suffixes and same val")

    options, args = parser.parse_args()
    if not args:
        parser.error("filename required")

    if options.debug:
        logging.basicConfig(level=logging.DEBUG)

    expand = []
    if options.expand:
        expand = [x.strip() for x in options.expand.split(",")]
    c = Parser(args[0], defaults=options.defaults, expand_defaults=expand,
               debug=options.debug)
    for s in args[1:]:
        c.parse_string(s)

    if options.debug:
        c.node.dump(0, True)

    dicts = c.get_dicts()
    print_dicts(options, dicts)
